# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for yapf.subtype_assigner."""

import textwrap
import unittest

from yapf.pytree import pytree_utils
from yapf.yapflib import format_token
from yapf.yapflib import subtypes

from yapftests import yapf_test_helper


class SubtypeAssignerTest(yapf_test_helper.YAPFTest):

  def _CheckFormatTokenSubtypes(self, llines, list_of_expected):
    """Check that the tokens in the LogicalLines have the expected subtypes.

    Args:
      llines: list of LogicalLine.
      list_of_expected: list of (name, subtype) pairs. Non-semantic tokens are
        filtered out from the expected values.
    """
    actual = []
    for lline in llines:
      filtered_values = [(ft.value, ft.subtypes)
                         for ft in lline.tokens
                         if ft.name not in pytree_utils.NONSEMANTIC_TOKENS]
      if filtered_values:
        actual.append(filtered_values)

    self.assertEqual(list_of_expected, actual)

  def testFuncDefDefaultAssign(self):
    self.maxDiff = None  # pylint: disable=invalid-name
    code = textwrap.dedent("""\
        def foo(a=37, *b, **c):
          return -x[:42]
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('a', {
                subtypes.NONE,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
                subtypes.PARAMETER_START,
            }),
            ('=', {
                subtypes.DEFAULT_OR_NAMED_ASSIGN,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            ('37', {
                subtypes.NONE,
                subtypes.PARAMETER_STOP,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            (',', {subtypes.NONE}),
            ('*', {
                subtypes.PARAMETER_START,
                subtypes.VARARGS_STAR,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            ('b', {
                subtypes.NONE,
                subtypes.PARAMETER_STOP,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            (',', {subtypes.NONE}),
            ('**', {
                subtypes.PARAMETER_START,
                subtypes.KWARGS_STAR_STAR,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            ('c', {
                subtypes.NONE,
                subtypes.PARAMETER_STOP,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('-', {subtypes.UNARY_OPERATOR}),
            ('x', {subtypes.NONE}),
            ('[', {subtypes.SUBSCRIPT_BRACKET}),
            (':', {subtypes.SUBSCRIPT_COLON}),
            ('42', {subtypes.NONE}),
            (']', {subtypes.SUBSCRIPT_BRACKET}),
        ],
    ])

  def testFuncCallWithDefaultAssign(self):
    code = textwrap.dedent("""\
        foo(x, a='hello world')
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('foo', {subtypes.NONE}),
            ('(', {subtypes.NONE}),
            ('x', {
                subtypes.NONE,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            (',', {subtypes.NONE}),
            ('a', {
                subtypes.NONE,
                subtypes.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST,
            }),
            ('=', {subtypes.DEFAULT_OR_NAMED_ASSIGN}),
            ("'hello world'", {subtypes.NONE}),
            (')', {subtypes.NONE}),
        ],
    ])

  def testSetComprehension(self):
    code = textwrap.dedent("""\
        def foo(value):
          return {value.lower()}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('value', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('value', {subtypes.NONE}),
            ('.', {subtypes.NONE}),
            ('lower', {subtypes.NONE}),
            ('(', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('}', {subtypes.NONE}),
        ],
    ])

    code = textwrap.dedent("""\
        def foo(strs):
          return {s.lower() for s in strs}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('strs', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('s', {subtypes.COMP_EXPR}),
            ('.', {subtypes.COMP_EXPR}),
            ('lower', {subtypes.COMP_EXPR}),
            ('(', {subtypes.COMP_EXPR}),
            (')', {subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
            }),
            ('s', {subtypes.COMP_FOR}),
            ('in', {subtypes.COMP_FOR}),
            ('strs', {subtypes.COMP_FOR}),
            ('}', {subtypes.NONE}),
        ],
    ])

    code = textwrap.dedent("""\
        def foo(strs):
          return {s + s.lower() for s in strs}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('strs', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('s', {subtypes.COMP_EXPR}),
            ('+', {subtypes.BINARY_OPERATOR, subtypes.COMP_EXPR}),
            ('s', {subtypes.COMP_EXPR}),
            ('.', {subtypes.COMP_EXPR}),
            ('lower', {subtypes.COMP_EXPR}),
            ('(', {subtypes.COMP_EXPR}),
            (')', {subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
            }),
            ('s', {subtypes.COMP_FOR}),
            ('in', {subtypes.COMP_FOR}),
            ('strs', {subtypes.COMP_FOR}),
            ('}', {subtypes.NONE}),
        ],
    ])

    code = textwrap.dedent("""\
        def foo(strs):
          return {c.lower() for s in strs for c in s}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('strs', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('c', {subtypes.COMP_EXPR}),
            ('.', {subtypes.COMP_EXPR}),
            ('lower', {subtypes.COMP_EXPR}),
            ('(', {subtypes.COMP_EXPR}),
            (')', {subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
                subtypes.COMP_EXPR,
            }),
            ('s', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('in', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('strs', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
            }),
            ('c', {subtypes.COMP_FOR}),
            ('in', {subtypes.COMP_FOR}),
            ('s', {subtypes.COMP_FOR}),
            ('}', {subtypes.NONE}),
        ],
    ])

  def testDictComprehension(self):
    code = textwrap.dedent("""\
        def foo(value):
          return {value: value.lower()}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('value', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('value', {subtypes.DICTIONARY_KEY, subtypes.DICTIONARY_KEY_PART}),
            (':', {subtypes.NONE}),
            ('value', {subtypes.DICTIONARY_VALUE}),
            ('.', {subtypes.NONE}),
            ('lower', {subtypes.NONE}),
            ('(', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('}', {subtypes.NONE}),
        ],
    ])

    code = textwrap.dedent("""\
        def foo(strs):
          return {s: s.lower() for s in strs}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('strs', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('s', {
                subtypes.DICTIONARY_KEY, subtypes.DICTIONARY_KEY_PART,
                subtypes.COMP_EXPR
            }),
            (':', {subtypes.COMP_EXPR}),
            ('s', {subtypes.DICTIONARY_VALUE, subtypes.COMP_EXPR}),
            ('.', {subtypes.COMP_EXPR}),
            ('lower', {subtypes.COMP_EXPR}),
            ('(', {subtypes.COMP_EXPR}),
            (')', {subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
            }),
            ('s', {subtypes.COMP_FOR}),
            ('in', {subtypes.COMP_FOR}),
            ('strs', {subtypes.COMP_FOR}),
            ('}', {subtypes.NONE}),
        ],
    ])

    code = textwrap.dedent("""\
        def foo(strs):
          return {c: c.lower() for s in strs for c in s}
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('def', {subtypes.NONE}),
            ('foo', {subtypes.FUNC_DEF}),
            ('(', {subtypes.NONE}),
            ('strs', {
                subtypes.NONE,
                subtypes.PARAMETER_START,
                subtypes.PARAMETER_STOP,
            }),
            (')', {subtypes.NONE}),
            (':', {subtypes.NONE}),
        ],
        [
            ('return', {subtypes.NONE}),
            ('{', {subtypes.NONE}),
            ('c', {
                subtypes.DICTIONARY_KEY, subtypes.DICTIONARY_KEY_PART,
                subtypes.COMP_EXPR
            }),
            (':', {subtypes.COMP_EXPR}),
            ('c', {subtypes.DICTIONARY_VALUE, subtypes.COMP_EXPR}),
            ('.', {subtypes.COMP_EXPR}),
            ('lower', {subtypes.COMP_EXPR}),
            ('(', {subtypes.COMP_EXPR}),
            (')', {subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
                subtypes.COMP_EXPR,
            }),
            ('s', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('in', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('strs', {subtypes.COMP_FOR, subtypes.COMP_EXPR}),
            ('for', {
                subtypes.DICT_SET_GENERATOR,
                subtypes.COMP_FOR,
            }),
            ('c', {subtypes.COMP_FOR}),
            ('in', {subtypes.COMP_FOR}),
            ('s', {subtypes.COMP_FOR}),
            ('}', {subtypes.NONE}),
        ],
    ])

  def testUnaryNotOperator(self):
    code = textwrap.dedent("""\
        not a
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [[('not', {subtypes.UNARY_OPERATOR}),
                                             ('a', {subtypes.NONE})]])

  def testBitwiseOperators(self):
    code = textwrap.dedent("""\
        x = ((a | (b ^ 3) & c) << 3) >> 1
        """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('x', {subtypes.NONE}),
            ('=', {subtypes.ASSIGN_OPERATOR}),
            ('(', {subtypes.NONE}),
            ('(', {subtypes.NONE}),
            ('a', {subtypes.NONE}),
            ('|', {subtypes.BINARY_OPERATOR}),
            ('(', {subtypes.NONE}),
            ('b', {subtypes.NONE}),
            ('^', {subtypes.BINARY_OPERATOR}),
            ('3', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('&', {subtypes.BINARY_OPERATOR}),
            ('c', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('<<', {subtypes.BINARY_OPERATOR}),
            ('3', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('>>', {subtypes.BINARY_OPERATOR}),
            ('1', {subtypes.NONE}),
        ],
    ])

  def testArithmeticOperators(self):
    code = textwrap.dedent("""\
        x = ((a + (b - 3) * (1 % c) @ d) / 3) // 1
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('x', {subtypes.NONE}),
            ('=', {subtypes.ASSIGN_OPERATOR}),
            ('(', {subtypes.NONE}),
            ('(', {subtypes.NONE}),
            ('a', {subtypes.NONE}),
            ('+', {subtypes.BINARY_OPERATOR}),
            ('(', {subtypes.NONE}),
            ('b', {subtypes.NONE}),
            ('-', {
                subtypes.BINARY_OPERATOR,
                subtypes.SIMPLE_EXPRESSION,
            }),
            ('3', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('*', {subtypes.BINARY_OPERATOR}),
            ('(', {subtypes.NONE}),
            ('1', {subtypes.NONE}),
            ('%', {
                subtypes.BINARY_OPERATOR,
                subtypes.SIMPLE_EXPRESSION,
            }),
            ('c', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('@', {subtypes.BINARY_OPERATOR}),
            ('d', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('/', {subtypes.BINARY_OPERATOR}),
            ('3', {subtypes.NONE}),
            (')', {subtypes.NONE}),
            ('//', {subtypes.BINARY_OPERATOR}),
            ('1', {subtypes.NONE}),
        ],
    ])

  def testSubscriptColon(self):
    code = textwrap.dedent("""\
        x[0:42:1]
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('x', {subtypes.NONE}),
            ('[', {subtypes.SUBSCRIPT_BRACKET}),
            ('0', {subtypes.NONE}),
            (':', {subtypes.SUBSCRIPT_COLON}),
            ('42', {subtypes.NONE}),
            (':', {subtypes.SUBSCRIPT_COLON}),
            ('1', {subtypes.NONE}),
            (']', {subtypes.SUBSCRIPT_BRACKET}),
        ],
    ])

  def testFunctionCallWithStarExpression(self):
    code = textwrap.dedent("""\
        [a, *b]
    """)
    llines = yapf_test_helper.ParseAndUnwrap(code)
    self._CheckFormatTokenSubtypes(llines, [
        [
            ('[', {subtypes.NONE}),
            ('a', {subtypes.NONE}),
            (',', {subtypes.NONE}),
            ('*', {
                subtypes.UNARY_OPERATOR,
                subtypes.VARARGS_STAR,
            }),
            ('b', {subtypes.NONE}),
            (']', {subtypes.NONE}),
        ],
    ])


if __name__ == '__main__':
  unittest.main()



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
"""Parser engine for the grammar tables generated by pgen.

The grammar table must be loaded first.

See Parser/parser.c in the Python distribution for additional info on
how this parsing engine works.

"""
from contextlib import contextmanager
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Text
from typing import Tuple
from typing import cast

from ..pytree import Context
from ..pytree import RawNode
from ..pytree import convert
# Local imports
from . import grammar
from . import token
from . import tokenize

DFA = List[List[Tuple[int, int]]]
DFAS = Tuple[DFA, Dict[int, int]]

# A placeholder node, used when parser is backtracking.
DUMMY_NODE = (-1, None, None, None)


def stack_copy(
    stack: List[Tuple[DFAS, int, RawNode]]) -> List[Tuple[DFAS, int, RawNode]]:
  """Nodeless stack copy."""
  return [(dfa, label, DUMMY_NODE) for dfa, label, _ in stack]


class Recorder:

  def __init__(self, parser: 'Parser', ilabels: List[int],
               context: Context) -> None:
    self.parser = parser
    self._ilabels = ilabels
    self.context = context  # not really matter

    self._dead_ilabels: Set[int] = set()
    self._start_point = self.parser.stack
    self._points = {ilabel: stack_copy(self._start_point) for ilabel in ilabels}

  @property
  def ilabels(self) -> Set[int]:
    return self._dead_ilabels.symmetric_difference(self._ilabels)

  @contextmanager
  def switch_to(self, ilabel: int) -> Iterator[None]:
    with self.backtrack():
      self.parser.stack = self._points[ilabel]
      try:
        yield
      except ParseError:
        self._dead_ilabels.add(ilabel)
      finally:
        self.parser.stack = self._start_point

  @contextmanager
  def backtrack(self) -> Iterator[None]:
    """
        Use the node-level invariant ones for basic parsing operations (push/pop/shift).
        These still will operate on the stack; but they won't create any new nodes, or
        modify the contents of any other existing nodes.
        This saves us a ton of time when we are backtracking, since we
        want to restore to the initial state as quick as possible, which
        can only be done by having as little mutatations as possible.
        """  # noqa: E501
    is_backtracking = self.parser.is_backtracking
    try:
      self.parser.is_backtracking = True
      yield
    finally:
      self.parser.is_backtracking = is_backtracking

  def add_token(self, tok_type: int, tok_val: Text, raw: bool = False) -> None:
    func: Callable[..., Any]
    if raw:
      func = self.parser._addtoken
    else:
      func = self.parser.addtoken

    for ilabel in self.ilabels:
      with self.switch_to(ilabel):
        args = [tok_type, tok_val, self.context]
        if raw:
          args.insert(0, ilabel)
        func(*args)

  def determine_route(self,
                      value: Text = None,
                      force: bool = False) -> Optional[int]:
    alive_ilabels = self.ilabels
    if len(alive_ilabels) == 0:
      *_, most_successful_ilabel = self._dead_ilabels
      raise ParseError('bad input', most_successful_ilabel, value, self.context)

    ilabel, *rest = alive_ilabels
    if force or not rest:
      return ilabel
    else:
      return None


class ParseError(Exception):
  """Exception to signal the parser is stuck."""

  def __init__(self, msg, type, value, context):
    Exception.__init__(
        self, '%s: type=%r, value=%r, context=%r' % (msg, type, value, context))
    self.msg = msg
    self.type = type
    self.value = value
    self.context = context

  def __reduce__(self):
    return type(self), (self.msg, self.type, self.value, self.context)


class Parser(object):
  """Parser engine.

    The proper usage sequence is:

    p = Parser(grammar, [converter])  # create instance
    p.setup([start])                  # prepare for parsing
    <for each input token>:
        if p.addtoken(...):           # parse a token; may raise ParseError
            break
    root = p.rootnode                 # root of abstract syntax tree

    A Parser instance may be reused by calling setup() repeatedly.

    A Parser instance contains state pertaining to the current token
    sequence, and should not be used concurrently by different threads
    to parse separate token sequences.

    See driver.py for how to get input tokens by tokenizing a file or
    string.

    Parsing is complete when addtoken() returns True; the root of the
    abstract syntax tree can then be retrieved from the rootnode
    instance variable.  When a syntax error occurs, addtoken() raises
    the ParseError exception.  There is no error recovery; the parser
    cannot be used after a syntax error was reported (but it can be
    reinitialized by calling setup()).

    """

  def __init__(self, grammar, convert=None):
    """Constructor.

        The grammar argument is a grammar.Grammar instance; see the
        grammar module for more information.

        The parser is not ready yet for parsing; you must call the
        setup() method to get it started.

        The optional convert argument is a function mapping concrete
        syntax tree nodes to abstract syntax tree nodes.  If not
        given, no conversion is done and the syntax tree produced is
        the concrete syntax tree.  If given, it must be a function of
        two arguments, the first being the grammar (a grammar.Grammar
        instance), and the second being the concrete syntax tree node
        to be converted.  The syntax tree is converted from the bottom
        up.

        A concrete syntax tree node is a (type, value, context, nodes)
        tuple, where type is the node type (a token or symbol number),
        value is None for symbols and a string for tokens, context is
        None or an opaque value used for error reporting (typically a
        (lineno, offset) pair), and nodes is a list of children for
        symbols, and None for tokens.

        An abstract syntax tree node may be anything; this is entirely
        up to the converter function.

        """
    self.grammar = grammar
    self.convert = convert or (lambda grammar, node: node)
    self.is_backtracking = False

  def setup(self, proxy, start=None):
    """Prepare for parsing.

        This *must* be called before starting to parse.

        The optional argument is an alternative start symbol; it
        defaults to the grammar's start symbol.

        You can use a Parser instance to parse any number of programs;
        each time you call setup() the parser is reset to an initial
        state determined by the (implicit or explicit) start symbol.

        """
    if start is None:
      start = self.grammar.start
    # Each stack entry is a tuple: (dfa, state, node).
    # A node is a tuple: (type, value, context, children),
    # where children is a list of nodes or None, and context may be None.
    newnode = (start, None, None, [])
    stackentry = (self.grammar.dfas[start], 0, newnode)
    self.stack = [stackentry]
    self.rootnode = None
    self.used_names = set()  # Aliased to self.rootnode.used_names in pop()
    self.proxy = proxy

  def addtoken(self, type, value, context):
    """Add a token; return True iff this is the end of the program."""
    # Map from token to label
    ilabels = self.classify(type, value, context)
    assert len(ilabels) >= 1

    # If we have only one state to advance, we'll directly
    # take it as is.
    if len(ilabels) == 1:
      [ilabel] = ilabels
      return self._addtoken(ilabel, type, value, context)

    # If there are multiple states which we can advance (only
    # happen under soft-keywords), then we will try all of them
    # in parallel and as soon as one state can reach further than
    # the rest, we'll choose that one. This is a pretty hacky
    # and hopefully temporary algorithm.
    #
    # For a more detailed explanation, check out this post:
    # https://tree.science/what-the-backtracking.html

    with self.proxy.release() as proxy:
      counter, force = 0, False
      recorder = Recorder(self, ilabels, context)
      recorder.add_token(type, value, raw=True)

      next_token_value = value
      while recorder.determine_route(next_token_value) is None:
        if not proxy.can_advance(counter):
          force = True
          break

        next_token_type, next_token_value, *_ = proxy.eat(counter)
        if next_token_type in (tokenize.COMMENT, tokenize.NL):
          counter += 1
          continue

        if next_token_type == tokenize.OP:
          next_token_type = grammar.opmap[next_token_value]

        recorder.add_token(next_token_type, next_token_value)
        counter += 1

      ilabel = cast(int,
                    recorder.determine_route(next_token_value, force=force))
      assert ilabel is not None

    return self._addtoken(ilabel, type, value, context)

  def _addtoken(self, ilabel: int, type: int, value: Text,
                context: Context) -> bool:
    # Loop until the token is shifted; may raise exceptions
    while True:
      dfa, state, node = self.stack[-1]
      states, first = dfa
      arcs = states[state]
      # Look for a state with this label
      for i, newstate in arcs:
        t = self.grammar.labels[i][0]
        if t >= 256:
          # See if it's a symbol and if we're in its first set
          itsdfa = self.grammar.dfas[t]
          itsstates, itsfirst = itsdfa
          if ilabel in itsfirst:
            # Push a symbol
            self.push(t, itsdfa, newstate, context)
            break  # To continue the outer while loop

        elif ilabel == i:
          # Look it up in the list of labels
          # Shift a token; we're done with it
          self.shift(type, value, newstate, context)
          # Pop while we are in an accept-only state
          state = newstate
          while states[state] == [(0, state)]:
            self.pop()
            if not self.stack:
              # Done parsing!
              return True
            dfa, state, node = self.stack[-1]
            states, first = dfa
          # Done with this token
          return False

      else:
        if (0, state) in arcs:
          # An accepting state, pop it and try something else
          self.pop()
          if not self.stack:
            # Done parsing, but another token is input
            raise ParseError('too much input', type, value, context)
        else:
          # No success finding a transition
          raise ParseError('bad input', type, value, context)

  def classify(self, type, value, context):
    """Turn a token into a label.  (Internal)

        Depending on whether the value is a soft-keyword or not,
        this function may return multiple labels to choose from."""
    if type == token.NAME:
      # Keep a listing of all used names
      self.used_names.add(value)
      # Check for reserved words
      if value in self.grammar.keywords:
        return [self.grammar.keywords[value]]
      elif value in self.grammar.soft_keywords:
        assert type in self.grammar.tokens
        return [
            self.grammar.soft_keywords[value],
            self.grammar.tokens[type],
        ]

    ilabel = self.grammar.tokens.get(type)
    if ilabel is None:
      raise ParseError('bad token', type, value, context)
    return [ilabel]

  def shift(self, type: int, value: Text, newstate: int,
            context: Context) -> None:
    """Shift a token.  (Internal)"""
    if self.is_backtracking:
      dfa, state, _ = self.stack[-1]
      self.stack[-1] = (dfa, newstate, DUMMY_NODE)
    else:
      dfa, state, node = self.stack[-1]
      rawnode: RawNode = (type, value, context, None)
      newnode = convert(self.grammar, rawnode)
      assert node[-1] is not None
      node[-1].append(newnode)
      self.stack[-1] = (dfa, newstate, node)

  def push(self, type: int, newdfa: DFAS, newstate: int,
           context: Context) -> None:
    """Push a nonterminal.  (Internal)"""
    if self.is_backtracking:
      dfa, state, _ = self.stack[-1]
      self.stack[-1] = (dfa, newstate, DUMMY_NODE)
      self.stack.append((newdfa, 0, DUMMY_NODE))
    else:
      dfa, state, node = self.stack[-1]
      newnode: RawNode = (type, None, context, [])
      self.stack[-1] = (dfa, newstate, node)
      self.stack.append((newdfa, 0, newnode))

  def pop(self) -> None:
    """Pop a nonterminal.  (Internal)"""
    if self.is_backtracking:
      self.stack.pop()
    else:
      popdfa, popstate, popnode = self.stack.pop()
      newnode = convert(self.grammar, popnode)
      if self.stack:
        dfa, state, node = self.stack[-1]
        assert node[-1] is not None
        node[-1].append(newnode)
      else:
        self.rootnode = newnode
        self.rootnode.used_names = self.used_names



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

# Modifications:
# Copyright 2006 Google, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
"""Parser driver.

This provides a high-level interface to parse a file into a syntax tree.

"""

__author__ = 'Guido van Rossum <guido@python.org>'

__all__ = ['Driver', 'load_grammar']

import io
import logging
import os
import pkgutil
import sys
# Python imports
from contextlib import contextmanager
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional

from platformdirs import user_cache_dir

from yapf._version import __version__ as yapf_version

# Pgen imports
from . import grammar
from . import parse
from . import pgen
from . import token
from . import tokenize


@dataclass
class ReleaseRange:
  start: int
  end: Optional[int] = None
  tokens: List[Any] = field(default_factory=list)

  def lock(self) -> None:
    total_eaten = len(self.tokens)
    self.end = self.start + total_eaten


class TokenProxy:

  def __init__(self, generator: Any) -> None:
    self._tokens = generator
    self._counter = 0
    self._release_ranges: List[ReleaseRange] = []

  @contextmanager
  def release(self) -> Iterator['TokenProxy']:
    release_range = ReleaseRange(self._counter)
    self._release_ranges.append(release_range)
    try:
      yield self
    finally:
      # Lock the last release range to the final position that
      # has been eaten.
      release_range.lock()

  def eat(self, point: int) -> Any:
    eaten_tokens = self._release_ranges[-1].tokens
    if point < len(eaten_tokens):
      return eaten_tokens[point]
    else:
      while point >= len(eaten_tokens):
        token = next(self._tokens)
        eaten_tokens.append(token)
      return token

  def __iter__(self) -> 'TokenProxy':
    return self

  def __next__(self) -> Any:
    # If the current position is already compromised (looked up)
    # return the eaten token, if not just go further on the given
    # token producer.
    for release_range in self._release_ranges:
      assert release_range.end is not None

      start, end = release_range.start, release_range.end
      if start <= self._counter < end:
        token = release_range.tokens[self._counter - start]
        break
    else:
      token = next(self._tokens)
    self._counter += 1
    return token

  def can_advance(self, to: int) -> bool:
    # Try to eat, fail if it can't. The eat operation is cached
    # so there wont be any additional cost of eating here
    try:
      self.eat(to)
    except StopIteration:
      return False
    else:
      return True


class Driver(object):

  def __init__(self, grammar, convert=None, logger=None):
    self.grammar = grammar
    if logger is None:
      logger = logging.getLogger()
    self.logger = logger
    self.convert = convert

  def parse_tokens(self, tokens, debug=False):
    """Parse a series of tokens and return the syntax tree."""
    # XXX Move the prefix computation into a wrapper around tokenize.
    p = parse.Parser(self.grammar, self.convert)
    proxy = TokenProxy(tokens)
    p.setup(proxy=proxy)
    lineno = 1
    column = 0
    type = value = start = end = line_text = None
    prefix = ''
    for quintuple in proxy:
      type, value, start, end, line_text = quintuple
      if start != (lineno, column):
        assert (lineno, column) <= start, ((lineno, column), start)
        s_lineno, s_column = start
        if lineno < s_lineno:
          prefix += '\n' * (s_lineno - lineno)
          lineno = s_lineno
          column = 0
        if column < s_column:
          prefix += line_text[column:s_column]
          column = s_column
      if type in (tokenize.COMMENT, tokenize.NL):
        prefix += value
        lineno, column = end
        if value.endswith('\n'):
          lineno += 1
          column = 0
        continue
      if type == token.OP:
        type = grammar.opmap[value]
      if debug:
        self.logger.debug('%s %r (prefix=%r)', token.tok_name[type], value,
                          prefix)
      if p.addtoken(type, value, (prefix, start)):
        if debug:
          self.logger.debug('Stop.')
        break
      prefix = ''
      lineno, column = end
      if value.endswith('\n'):
        lineno += 1
        column = 0
    else:
      # We never broke out -- EOF is too soon (how can this happen???)
      raise parse.ParseError('incomplete input', type, value, (prefix, start))
    return p.rootnode

  def parse_stream_raw(self, stream, debug=False):
    """Parse a stream and return the syntax tree."""
    tokens = tokenize.generate_tokens(stream.readline)
    return self.parse_tokens(tokens, debug)

  def parse_stream(self, stream, debug=False):
    """Parse a stream and return the syntax tree."""
    return self.parse_stream_raw(stream, debug)

  def parse_file(self, filename, encoding=None, debug=False):
    """Parse a file and return the syntax tree."""
    with io.open(filename, 'r', encoding=encoding) as stream:
      return self.parse_stream(stream, debug)

  def parse_string(self, text, debug=False):
    """Parse a string and return the syntax tree."""
    tokens = tokenize.generate_tokens(io.StringIO(text).readline)
    return self.parse_tokens(tokens, debug)


def _generate_pickle_name(gt):
  # type:(str) -> str
  """Get the filepath to write a pickle file to
  given the path of a grammar textfile.

  The returned filepath should be in a user-specific cache directory.

  Args:
      gt (str): path to grammar text file

  Returns:
      str: path to pickle file
  """

  grammar_textfile_name = os.path.basename(gt)
  head, tail = os.path.splitext(grammar_textfile_name)
  if tail == '.txt':
    tail = ''
  cache_dir = user_cache_dir(
      appname='YAPF', appauthor='Google', version=yapf_version)
  return cache_dir + os.sep + head + tail + '-py' + '.'.join(
      map(str, sys.version_info)) + '.pickle'


def load_grammar(gt='Grammar.txt',
                 gp=None,
                 save=True,
                 force=False,
                 logger=None):
  # type:(str, str | None, bool, bool, logging.Logger | None) -> grammar.Grammar
  """Load the grammar (maybe from a pickle)."""
  if logger is None:
    logger = logging.getLogger()
  gp = _generate_pickle_name(gt) if gp is None else gp
  grammar_text = gt
  try:
    newer = _newer(gp, gt)
  except OSError as err:
    logger.debug('OSError, could not check if newer: %s', err.args)
    newer = True
  if not os.path.exists(gt):
    # Assume package data
    gt_basename = os.path.basename(gt)
    pd = pkgutil.get_data('yapf_third_party._ylib2to3', gt_basename)
    if pd is None:
      raise RuntimeError('Failed to load grammer %s from package' % gt_basename)
    grammar_text = io.StringIO(pd.decode(encoding='utf-8'))
  if force or not newer:
    g = pgen.generate_grammar(grammar_text)
    if save:
      try:
        Path(gp).parent.mkdir(parents=True, exist_ok=True)
        g.dump(gp)
      except OSError:
        # Ignore error, caching is not vital.
        pass
  else:
    g = grammar.Grammar()
    g.load(gp)
  return g


def _newer(a, b):
  """Inquire whether file a was written since file b."""
  if not os.path.exists(a):
    return False
  if not os.path.exists(b):
    return True
  return os.path.getmtime(a) >= os.path.getmtime(b)


def load_packaged_grammar(package, grammar_source):
  """Normally, loads a pickled grammar by doing
        pkgutil.get_data(package, pickled_grammar)
    where *pickled_grammar* is computed from *grammar_source* by adding the
    Python version and using a ``.pickle`` extension.

    However, if *grammar_source* is an extant file, load_grammar(grammar_source)
    is called instead. This facilitates using a packaged grammar file when needed
    but preserves load_grammar's automatic regeneration behavior when possible.

    """  # noqa: E501
  if os.path.isfile(grammar_source):
    return load_grammar(grammar_source)
  pickled_name = _generate_pickle_name(os.path.basename(grammar_source))
  data = pkgutil.get_data(package, pickled_name)
  g = grammar.Grammar()
  g.loads(data)
  return g


def main(*args):
  """Main program, when run as a script: produce grammar pickle files.

    Calls load_grammar for each argument, a path to a grammar text file.
    """
  if not args:
    args = sys.argv[1:]
  logging.basicConfig(
      level=logging.INFO, stream=sys.stdout, format='%(message)s')
  for gt in args:
    load_grammar(gt, save=True, force=True)
  return True


if __name__ == '__main__':
  sys.exit(int(not main()))



# Copyright 2022 Bill Wendling, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyParser-related utilities.

This module collects various utilities related to the parse trees produced by
the pyparser.

  GetLogicalLine: produces a list of tokens from the logical lines within a
    range.
  GetTokensInSubRange: produces a sublist of tokens from a current token list
    within a range.
  GetTokenIndex: Get the index of a token.
  GetNextTokenIndex: Get the index of the next token after a given position.
  GetPrevTokenIndex: Get the index of the previous token before a given
    position.
  TokenStart: Convenience function to return the token's start as a tuple.
  TokenEnd: Convenience function to return the token's end as a tuple.
"""


def GetLogicalLine(logical_lines, node):
  """Get a list of tokens within the node's range from the logical lines."""
  start = TokenStart(node)
  end = TokenEnd(node)
  tokens = []

  for line in logical_lines:
    if line.start > end:
      break
    if line.start <= start or line.end >= end:
      tokens.extend(GetTokensInSubRange(line.tokens, node))

  return tokens


def GetTokensInSubRange(tokens, node):
  """Get a subset of tokens representing the node."""
  start = TokenStart(node)
  end = TokenEnd(node)
  tokens_in_range = []

  for tok in tokens:
    tok_range = (tok.lineno, tok.column)
    if tok_range >= start and tok_range < end:
      tokens_in_range.append(tok)

  return tokens_in_range


def GetTokenIndex(tokens, pos):
  """Get the index of the token at pos."""
  for index, token in enumerate(tokens):
    if (token.lineno, token.column) == pos:
      return index

  return None


def GetNextTokenIndex(tokens, pos):
  """Get the index of the next token after pos."""
  for index, token in enumerate(tokens):
    if (token.lineno, token.column) >= pos:
      return index

  return None


def GetPrevTokenIndex(tokens, pos):
  """Get the index of the previous token before pos."""
  for index, token in enumerate(tokens):
    if index > 0 and (token.lineno, token.column) >= pos:
      return index - 1

  return None


def TokenStart(node):
  return (node.lineno, node.col_offset)


def TokenEnd(node):
  return (node.end_lineno, node.end_col_offset)


#############################################################################
# Code for debugging                                                        #
#############################################################################


def AstDump(node):
  import ast
  print(ast.dump(node, include_attributes=True, indent=4))



# Modified copy of clang-format-diff.py that works with yapf.
#
# Licensed under the Apache License, Version 2.0 (the "License") with LLVM
# Exceptions; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://llvm.org/LICENSE.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script reads input from a unified diff and reformats all the changed
lines. This is useful to reformat all the lines touched by a specific patch.
Example usage for git/svn users:

  git diff -U0 --no-color --relative HEAD^ | yapf-diff -i
  svn diff --diff-cmd=diff -x-U0 | yapf-diff -p0 -i

It should be noted that the filename contained in the diff is used unmodified
to determine the source file to update. Users calling this script directly
should be careful to ensure that the path in the diff is correct relative to the
current working directory.
"""

import argparse
import difflib
import re
import subprocess
import sys
from io import StringIO


def main():
  parser = argparse.ArgumentParser(
      description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
  parser.add_argument(
      '-i',
      '--in-place',
      action='store_true',
      default=False,
      help='apply edits to files instead of displaying a diff')
  parser.add_argument(
      '-p',
      '--prefix',
      metavar='NUM',
      default=1,
      help='strip the smallest prefix containing P slashes')
  parser.add_argument(
      '--regex',
      metavar='PATTERN',
      default=None,
      help='custom pattern selecting file paths to reformat '
      '(case sensitive, overrides -iregex)')
  parser.add_argument(
      '--iregex',
      metavar='PATTERN',
      default=r'.*\.(py)',
      help='custom pattern selecting file paths to reformat '
      '(case insensitive, overridden by -regex)')
  parser.add_argument(
      '-v',
      '--verbose',
      action='store_true',
      help='be more verbose, ineffective without -i')
  parser.add_argument(
      '--style',
      help='specify formatting style: either a style name (for '
      'example "pep8" or "google"), or the name of a file with '
      'style settings. The default is pep8 unless a '
      '.style.yapf or setup.cfg file located in one of the '
      'parent directories of the source file (or current '
      'directory for stdin)')
  parser.add_argument(
      '--binary', default='yapf', help='location of binary to use for yapf')
  args = parser.parse_args()

  # Extract changed lines for each file.
  filename = None
  lines_by_file = {}
  for line in sys.stdin:
    match = re.search(r'^\+\+\+\ (.*?/){%s}(\S*)' % args.prefix, line)
    if match:
      filename = match.group(2)
    if filename is None:
      continue

    if args.regex is not None:
      if not re.match('^%s$' % args.regex, filename):
        continue
    elif not re.match('^%s$' % args.iregex, filename, re.IGNORECASE):
      continue

    match = re.search(r'^@@.*\+(\d+)(,(\d+))?', line)
    if match:
      start_line = int(match.group(1))
      line_count = 1
      if match.group(3):
        line_count = int(match.group(3))
      if line_count == 0:
        continue
      end_line = start_line + line_count - 1
      lines_by_file.setdefault(filename, []).extend(
          ['--lines', str(start_line) + '-' + str(end_line)])

  # Reformat files containing changes in place.
  for filename, lines in lines_by_file.items():
    if args.in_place and args.verbose:
      print('Formatting {}'.format(filename))
    command = [args.binary, filename]
    if args.in_place:
      command.append('-i')
    command.extend(lines)
    if args.style:
      command.extend(['--style', args.style])
    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=None,
        stdin=subprocess.PIPE,
        universal_newlines=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
      sys.exit(p.returncode)

    if not args.in_place:
      with open(filename) as f:
        code = f.readlines()
      formatted_code = StringIO(stdout).readlines()
      diff = difflib.unified_diff(code, formatted_code, filename, filename,
                                  '(before formatting)', '(after formatting)')
      diff_string = ''.join(diff)
      if len(diff_string) > 0:
        sys.stdout.write(diff_string)


if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for yapf.file_resources."""

import codecs
import contextlib
import os
import shutil
import tempfile
import unittest
from io import BytesIO

from yapf.yapflib import errors
from yapf.yapflib import file_resources

from yapftests import utils
from yapftests import yapf_test_helper


@contextlib.contextmanager
def _restore_working_dir():
  curdir = os.getcwd()
  try:
    yield
  finally:
    os.chdir(curdir)


@contextlib.contextmanager
def _exists_mocked_in_module(module, mock_implementation):
  unmocked_exists = getattr(module, 'exists')
  setattr(module, 'exists', mock_implementation)
  try:
    yield
  finally:
    setattr(module, 'exists', unmocked_exists)


class GetExcludePatternsForDir(yapf_test_helper.YAPFTest):

  def setUp(self):  # pylint: disable=g-missing-super-call
    self.test_tmpdir = tempfile.mkdtemp()

  def tearDown(self):  # pylint: disable=g-missing-super-call
    shutil.rmtree(self.test_tmpdir)

  def test_get_exclude_file_patterns_from_yapfignore(self):
    local_ignore_file = os.path.join(self.test_tmpdir, '.yapfignore')
    ignore_patterns = ['temp/**/*.py', 'temp2/*.py']
    with open(local_ignore_file, 'w') as f:
      f.writelines('\n'.join(ignore_patterns))

    self.assertEqual(
        sorted(file_resources.GetExcludePatternsForDir(self.test_tmpdir)),
        sorted(ignore_patterns))

  def test_get_exclude_file_patterns_from_yapfignore_with_wrong_syntax(self):
    local_ignore_file = os.path.join(self.test_tmpdir, '.yapfignore')
    ignore_patterns = ['temp/**/*.py', './wrong/syntax/*.py']
    with open(local_ignore_file, 'w') as f:
      f.writelines('\n'.join(ignore_patterns))

    with self.assertRaises(errors.YapfError):
      file_resources.GetExcludePatternsForDir(self.test_tmpdir)

  def test_get_exclude_file_patterns_from_pyproject(self):
    local_ignore_file = os.path.join(self.test_tmpdir, 'pyproject.toml')
    ignore_patterns = ['temp/**/*.py', 'temp2/*.py']
    with open(local_ignore_file, 'w') as f:
      f.write('[tool.yapfignore]\n')
      f.write('ignore_patterns=[')
      f.writelines('\n,'.join(['"{}"'.format(p) for p in ignore_patterns]))
      f.write(']')

    self.assertEqual(
        sorted(file_resources.GetExcludePatternsForDir(self.test_tmpdir)),
        sorted(ignore_patterns))

  def test_get_exclude_file_patterns_from_pyproject_no_ignore_section(self):
    local_ignore_file = os.path.join(self.test_tmpdir, 'pyproject.toml')
    ignore_patterns = []
    open(local_ignore_file, 'w').close()

    self.assertEqual(
        sorted(file_resources.GetExcludePatternsForDir(self.test_tmpdir)),
        sorted(ignore_patterns))

  def test_get_exclude_file_patterns_from_pyproject_ignore_section_empty(self):
    local_ignore_file = os.path.join(self.test_tmpdir, 'pyproject.toml')
    ignore_patterns = []
    with open(local_ignore_file, 'w') as f:
      f.write('[tool.yapfignore]\n')

    self.assertEqual(
        sorted(file_resources.GetExcludePatternsForDir(self.test_tmpdir)),
        sorted(ignore_patterns))

  def test_get_exclude_file_patterns_with_no_config_files(self):
    ignore_patterns = []

    self.assertEqual(
        sorted(file_resources.GetExcludePatternsForDir(self.test_tmpdir)),
        sorted(ignore_patterns))


class GetDefaultStyleForDirTest(yapf_test_helper.YAPFTest):

  def setUp(self):  # pylint: disable=g-missing-super-call
    self.test_tmpdir = tempfile.mkdtemp()

  def tearDown(self):  # pylint: disable=g-missing-super-call
    shutil.rmtree(self.test_tmpdir)

  def test_no_local_style(self):
    test_file = os.path.join(self.test_tmpdir, 'file.py')
    style_name = file_resources.GetDefaultStyleForDir(test_file)
    self.assertEqual(style_name, 'pep8')

  def test_no_local_style_custom_default(self):
    test_file = os.path.join(self.test_tmpdir, 'file.py')
    style_name = file_resources.GetDefaultStyleForDir(
        test_file, default_style='custom-default')
    self.assertEqual(style_name, 'custom-default')

  def test_with_local_style(self):
    # Create an empty .style.yapf file in test_tmpdir
    style_file = os.path.join(self.test_tmpdir, '.style.yapf')
    open(style_file, 'w').close()

    test_filename = os.path.join(self.test_tmpdir, 'file.py')
    self.assertEqual(style_file,
                     file_resources.GetDefaultStyleForDir(test_filename))

    test_filename = os.path.join(self.test_tmpdir, 'dir1', 'file.py')
    self.assertEqual(style_file,
                     file_resources.GetDefaultStyleForDir(test_filename))

  def test_setup_config(self):
    # An empty setup.cfg file should not be used
    setup_config = os.path.join(self.test_tmpdir, 'setup.cfg')
    open(setup_config, 'w').close()

    test_dir = os.path.join(self.test_tmpdir, 'dir1')
    style_name = file_resources.GetDefaultStyleForDir(test_dir)
    self.assertEqual(style_name, 'pep8')

    # One with a '[yapf]' section should be used
    with open(setup_config, 'w') as f:
      f.write('[yapf]\n')
    self.assertEqual(setup_config,
                     file_resources.GetDefaultStyleForDir(test_dir))

  def test_pyproject_toml(self):
    pyproject_toml = os.path.join(self.test_tmpdir, 'pyproject.toml')
    open(pyproject_toml, 'w').close()

    test_dir = os.path.join(self.test_tmpdir, 'dir1')
    style_name = file_resources.GetDefaultStyleForDir(test_dir)
    self.assertEqual(style_name, 'pep8')

    # One with a '[tool.yapf]' section should be used
    with open(pyproject_toml, 'w') as f:
      f.write('[tool.yapf]\n')
    self.assertEqual(pyproject_toml,
                     file_resources.GetDefaultStyleForDir(test_dir))

  def test_local_style_at_root(self):
    # Test behavior of files located on the root, and under root.
    rootdir = os.path.abspath(os.path.sep)
    test_dir_at_root = os.path.join(rootdir, 'dir1')
    test_dir_under_root = os.path.join(rootdir, 'dir1', 'dir2')

    # Fake placing only a style file at the root by mocking `os.path.exists`.
    style_file = os.path.join(rootdir, '.style.yapf')

    def mock_exists_implementation(path):
      return path == style_file

    with _exists_mocked_in_module(file_resources.os.path,
                                  mock_exists_implementation):
      # Both files should find the style file at the root.
      default_style_at_root = file_resources.GetDefaultStyleForDir(
          test_dir_at_root)
      self.assertEqual(style_file, default_style_at_root)
      default_style_under_root = file_resources.GetDefaultStyleForDir(
          test_dir_under_root)
      self.assertEqual(style_file, default_style_under_root)


def _touch_files(filenames):
  for name in filenames:
    open(name, 'a').close()


class GetCommandLineFilesTest(yapf_test_helper.YAPFTest):

  def setUp(self):  # pylint: disable=g-missing-super-call
    self.test_tmpdir = tempfile.mkdtemp()
    self.old_dir = os.getcwd()

  def tearDown(self):  # pylint: disable=g-missing-super-call
    os.chdir(self.old_dir)
    shutil.rmtree(self.test_tmpdir)

  def _make_test_dir(self, name):
    fullpath = os.path.normpath(os.path.join(self.test_tmpdir, name))
    os.makedirs(fullpath)
    return fullpath

  def test_find_files_not_dirs(self):
    tdir1 = self._make_test_dir('test1')
    tdir2 = self._make_test_dir('test2')
    file1 = os.path.join(tdir1, 'testfile1.py')
    file2 = os.path.join(tdir2, 'testfile2.py')
    _touch_files([file1, file2])

    self.assertEqual(
        file_resources.GetCommandLineFiles([file1, file2],
                                           recursive=False,
                                           exclude=None), [file1, file2])
    self.assertEqual(
        file_resources.GetCommandLineFiles([file1, file2],
                                           recursive=True,
                                           exclude=None), [file1, file2])

  def test_nonrecursive_find_in_dir(self):
    tdir1 = self._make_test_dir('test1')
    tdir2 = self._make_test_dir('test1/foo')
    file1 = os.path.join(tdir1, 'testfile1.py')
    file2 = os.path.join(tdir2, 'testfile2.py')
    _touch_files([file1, file2])

    self.assertRaises(
        errors.YapfError,
        file_resources.GetCommandLineFiles,
        command_line_file_list=[tdir1],
        recursive=False,
        exclude=None)

  def test_recursive_find_in_dir(self):
    tdir1 = self._make_test_dir('test1')
    tdir2 = self._make_test_dir('test2/testinner/')
    tdir3 = self._make_test_dir('test3/foo/bar/bas/xxx')
    files = [
        os.path.join(tdir1, 'testfile1.py'),
        os.path.join(tdir2, 'testfile2.py'),
        os.path.join(tdir3, 'testfile3.py'),
    ]
    _touch_files(files)

    self.assertEqual(
        sorted(
            file_resources.GetCommandLineFiles([self.test_tmpdir],
                                               recursive=True,
                                               exclude=None)), sorted(files))

  def test_recursive_find_in_dir_with_exclude(self):
    tdir1 = self._make_test_dir('test1')
    tdir2 = self._make_test_dir('test2/testinner/')
    tdir3 = self._make_test_dir('test3/foo/bar/bas/xxx')
    files = [
        os.path.join(tdir1, 'testfile1.py'),
        os.path.join(tdir2, 'testfile2.py'),
        os.path.join(tdir3, 'testfile3.py'),
    ]
    _touch_files(files)

    self.assertEqual(
        sorted(
            file_resources.GetCommandLineFiles([self.test_tmpdir],
                                               recursive=True,
                                               exclude=['*test*3.py'])),
        sorted([
            os.path.join(tdir1, 'testfile1.py'),
            os.path.join(tdir2, 'testfile2.py'),
        ]))

  def test_find_with_excluded_hidden_dirs(self):
    tdir1 = self._make_test_dir('.test1')
    tdir2 = self._make_test_dir('test_2')
    tdir3 = self._make_test_dir('test.3')
    files = [
        os.path.join(tdir1, 'testfile1.py'),
        os.path.join(tdir2, 'testfile2.py'),
        os.path.join(tdir3, 'testfile3.py'),
    ]
    _touch_files(files)

    actual = file_resources.GetCommandLineFiles([self.test_tmpdir],
                                                recursive=True,
                                                exclude=['*.test1*'])

    self.assertEqual(
        sorted(actual),
        sorted([
            os.path.join(tdir2, 'testfile2.py'),
            os.path.join(tdir3, 'testfile3.py'),
        ]))

  def test_find_with_excluded_hidden_dirs_relative(self):
    """Test find with excluded hidden dirs.

    A regression test against a specific case where a hidden directory (one
    beginning with a period) is being excluded, but it is also an immediate
    child of the current directory which has been specified in a relative
    manner.

    At its core, the bug has to do with overzealous stripping of "./foo" so that
    it removes too much from "./.foo" .
    """
    tdir1 = self._make_test_dir('.test1')
    tdir2 = self._make_test_dir('test_2')
    tdir3 = self._make_test_dir('test.3')
    files = [
        os.path.join(tdir1, 'testfile1.py'),
        os.path.join(tdir2, 'testfile2.py'),
        os.path.join(tdir3, 'testfile3.py'),
    ]
    _touch_files(files)

    # We must temporarily change the current directory, so that we test against
    # patterns like ./.test1/file instead of /tmp/foo/.test1/file
    with _restore_working_dir():

      os.chdir(self.test_tmpdir)
      actual = file_resources.GetCommandLineFiles(
          [os.path.relpath(self.test_tmpdir)],
          recursive=True,
          exclude=['*.test1*'])

      self.assertEqual(
          sorted(actual),
          sorted([
              os.path.join(
                  os.path.relpath(self.test_tmpdir), os.path.basename(tdir2),
                  'testfile2.py'),
              os.path.join(
                  os.path.relpath(self.test_tmpdir), os.path.basename(tdir3),
                  'testfile3.py'),
          ]))

  def test_find_with_excluded_dirs(self):
    tdir1 = self._make_test_dir('test1')
    tdir2 = self._make_test_dir('test2/testinner/')
    tdir3 = self._make_test_dir('test3/foo/bar/bas/xxx')
    files = [
        os.path.join(tdir1, 'testfile1.py'),
        os.path.join(tdir2, 'testfile2.py'),
        os.path.join(tdir3, 'testfile3.py'),
    ]
    _touch_files(files)

    os.chdir(self.test_tmpdir)

    found = sorted(
        file_resources.GetCommandLineFiles(['test1', 'test2', 'test3'],
                                           recursive=True,
                                           exclude=[
                                               'test1',
                                               'test2/testinner/',
                                           ]))

    self.assertEqual(
        found, ['test3/foo/bar/bas/xxx/testfile3.py'.replace('/', os.path.sep)])

    found = sorted(
        file_resources.GetCommandLineFiles(['.'],
                                           recursive=True,
                                           exclude=[
                                               'test1',
                                               'test3',
                                           ]))

    self.assertEqual(
        found, ['./test2/testinner/testfile2.py'.replace('/', os.path.sep)])

  def test_find_with_excluded_current_dir(self):
    with self.assertRaises(errors.YapfError):
      file_resources.GetCommandLineFiles([], False, exclude=['./z'])


class IsPythonFileTest(yapf_test_helper.YAPFTest):

  def setUp(self):  # pylint: disable=g-missing-super-call
    self.test_tmpdir = tempfile.mkdtemp()

  def tearDown(self):  # pylint: disable=g-missing-super-call
    shutil.rmtree(self.test_tmpdir)

  def test_with_py_extension(self):
    file1 = os.path.join(self.test_tmpdir, 'testfile1.py')
    self.assertTrue(file_resources.IsPythonFile(file1))

  def test_empty_without_py_extension(self):
    file1 = os.path.join(self.test_tmpdir, 'testfile1')
    self.assertFalse(file_resources.IsPythonFile(file1))
    file2 = os.path.join(self.test_tmpdir, 'testfile1.rb')
    self.assertFalse(file_resources.IsPythonFile(file2))

  def test_python_shebang(self):
    file1 = os.path.join(self.test_tmpdir, 'testfile1')
    with open(file1, 'w') as f:
      f.write('#!/usr/bin/python\n')
    self.assertTrue(file_resources.IsPythonFile(file1))

    file2 = os.path.join(self.test_tmpdir, 'testfile2.run')
    with open(file2, 'w') as f:
      f.write('#! /bin/python2\n')
    self.assertTrue(file_resources.IsPythonFile(file1))

  def test_with_latin_encoding(self):
    file1 = os.path.join(self.test_tmpdir, 'testfile1')
    with codecs.open(file1, mode='w', encoding='latin-1') as f:
      f.write('#! /bin/python2\n')
    self.assertTrue(file_resources.IsPythonFile(file1))

  def test_with_invalid_encoding(self):
    file1 = os.path.join(self.test_tmpdir, 'testfile1')
    with open(file1, 'w') as f:
      f.write('#! /bin/python2\n')
      f.write('# -*- coding: iso-3-14159 -*-\n')
    self.assertFalse(file_resources.IsPythonFile(file1))


class IsIgnoredTest(yapf_test_helper.YAPFTest):

  def test_root_path(self):
    self.assertTrue(file_resources.IsIgnored('media', ['media']))
    self.assertFalse(file_resources.IsIgnored('media', ['media/*']))

  def test_sub_path(self):
    self.assertTrue(file_resources.IsIgnored('media/a', ['*/a']))
    self.assertTrue(file_resources.IsIgnored('media/b', ['media/*']))
    self.assertTrue(file_resources.IsIgnored('media/b/c', ['*/*/c']))

  def test_trailing_slash(self):
    self.assertTrue(file_resources.IsIgnored('z', ['z']))
    self.assertTrue(file_resources.IsIgnored('z', ['z' + os.path.sep]))


class BufferedByteStream(object):

  def __init__(self):
    self.stream = BytesIO()

  def getvalue(self):  # pylint: disable=invalid-name
    return self.stream.getvalue().decode('utf-8')

  @property
  def buffer(self):
    return self.stream


class WriteReformattedCodeTest(yapf_test_helper.YAPFTest):

  @classmethod
  def setUpClass(cls):  # pylint: disable=g-missing-super-call
    cls.test_tmpdir = tempfile.mkdtemp()

  @classmethod
  def tearDownClass(cls):  # pylint: disable=g-missing-super-call
    shutil.rmtree(cls.test_tmpdir)

  def test_write_to_file(self):
    s = 'foobar\n'
    with utils.NamedTempFile(dirname=self.test_tmpdir) as (f, fname):
      file_resources.WriteReformattedCode(
          fname, s, in_place=True, encoding='utf-8')
      f.flush()

      with open(fname) as f2:
        self.assertEqual(f2.read(), s)

  def test_write_to_stdout(self):
    s = 'foobar'
    stream = BufferedByteStream()
    with utils.stdout_redirector(stream):
      file_resources.WriteReformattedCode(
          None, s, in_place=False, encoding='utf-8')
    self.assertEqual(stream.getvalue(), s)

  def test_write_encoded_to_stdout(self):
    s = '\ufeff# -*- coding: utf-8 -*-\nresult = "passed"\n'  # pylint: disable=anomalous-unicode-escape-in-string # noqa
    stream = BufferedByteStream()
    with utils.stdout_redirector(stream):
      file_resources.WriteReformattedCode(
          None, s, in_place=False, encoding='utf-8')
    self.assertEqual(stream.getvalue(), s)


class LineEndingTest(yapf_test_helper.YAPFTest):

  def test_line_ending_linefeed(self):
    lines = ['spam\n', 'spam\n']
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\n')

  def test_line_ending_carriage_return(self):
    lines = ['spam\r', 'spam\r']
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\r')

  def test_line_ending_combo(self):
    lines = ['spam\r\n', 'spam\r\n']
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\r\n')

  def test_line_ending_weighted(self):
    lines = [
        'spam\n',
        'spam\n',
        'spam\r',
        'spam\r\n',
    ]
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\n')

  def test_line_ending_empty(self):
    lines = []
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\n')

  def test_line_ending_no_newline(self):
    lines = ['spam']
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\n')

  def test_line_ending_tie(self):
    lines = [
        'spam\n',
        'spam\n',
        'spam\r\n',
        'spam\r\n',
    ]
    actual = file_resources.LineEnding(lines)
    self.assertEqual(actual, '\n')


if __name__ == '__main__':
  unittest.main()



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Enhanced token information for formatting."""

import keyword
import re
from functools import lru_cache

from yapf_third_party._ylib2to3.pgen2 import token
from yapf_third_party._ylib2to3.pytree import type_repr

from yapf.pytree import pytree_utils
from yapf.yapflib import style
from yapf.yapflib import subtypes

CONTINUATION = token.N_TOKENS

_OPENING_BRACKETS = frozenset({'(', '[', '{'})
_CLOSING_BRACKETS = frozenset({')', ']', '}'})


def _TabbedContinuationAlignPadding(spaces, align_style, tab_width):
  """Build padding string for continuation alignment in tabbed indentation.

  Arguments:
    spaces: (int) The number of spaces to place before the token for alignment.
    align_style: (str) The alignment style for continuation lines.
    tab_width: (int) Number of columns of each tab character.

  Returns:
    A padding string for alignment with style specified by align_style option.
  """
  if align_style in ('FIXED', 'VALIGN-RIGHT'):
    if spaces > 0:
      return '\t' * int((spaces + tab_width - 1) / tab_width)
    return ''
  return ' ' * spaces


class FormatToken(object):
  """Enhanced token information for formatting.

  This represents the token plus additional information useful for reformatting
  the code.

  Attributes:
    node: The original token node.
    next_token: The token in the logical line after this token or None if this
      is the last token in the logical line.
    previous_token: The token in the logical line before this token or None if
      this is the first token in the logical line.
    matching_bracket: If a bracket token ('[', '{', or '(') the matching
      bracket.
    parameters: If this and its following tokens make up a parameter list, then
      this is a list of those parameters.
    container_opening: If the object is in a container, this points to its
      opening bracket.
    container_elements: If this is the start of a container, a list of the
      elements in the container.
    whitespace_prefix: The prefix for the whitespace.
    spaces_required_before: The number of spaces required before a token. This
      is a lower-bound for the formatter and not a hard requirement. For
      instance, a comment may have n required spaces before it. But the
      formatter won't place n spaces before all comments. Only those that are
      moved to the end of a line of code. The formatter may use different
      spacing when appropriate.
    total_length: The total length of the logical line up to and including
      whitespace and this token. However, this doesn't include the initial
      indentation amount.
    split_penalty: The penalty for splitting the line before this token.
    can_break_before: True if we're allowed to break before this token.
    must_break_before: True if we're required to break before this token.
    newlines: The number of newlines needed before this token.
  """

  def __init__(self, node, name):
    """Constructor.

    Arguments:
      node: (pytree.Leaf) The node that's being wrapped.
      name: (string) The name of the node.
    """
    self.node = node
    self.name = name
    self.type = node.type
    self.column = node.column
    self.lineno = node.lineno
    self.value = node.value

    if self.is_continuation:
      self.value = node.value.rstrip()

    self.next_token = None
    self.previous_token = None
    self.matching_bracket = None
    self.parameters = []
    self.container_opening = None
    self.container_elements = []
    self.whitespace_prefix = ''
    self.total_length = 0
    self.split_penalty = 0
    self.can_break_before = False
    self.must_break_before = pytree_utils.GetNodeAnnotation(
        node, pytree_utils.Annotation.MUST_SPLIT, default=False)
    self.newlines = pytree_utils.GetNodeAnnotation(
        node, pytree_utils.Annotation.NEWLINES)
    self.spaces_required_before = 0

    if self.is_comment:
      self.spaces_required_before = style.Get('SPACES_BEFORE_COMMENT')

    stypes = pytree_utils.GetNodeAnnotation(node,
                                            pytree_utils.Annotation.SUBTYPE)
    self.subtypes = {subtypes.NONE} if not stypes else stypes
    self.is_pseudo = hasattr(node, 'is_pseudo') and node.is_pseudo

  @property
  def formatted_whitespace_prefix(self):
    if style.Get('INDENT_BLANK_LINES'):
      without_newlines = self.whitespace_prefix.lstrip('\n')
      height = len(self.whitespace_prefix) - len(without_newlines)
      if height:
        return ('\n' + without_newlines) * height
    return self.whitespace_prefix

  def AddWhitespacePrefix(self, newlines_before, spaces=0, indent_level=0):
    """Register a token's whitespace prefix.

    This is the whitespace that will be output before a token's string.

    Arguments:
      newlines_before: (int) The number of newlines to place before the token.
      spaces: (int) The number of spaces to place before the token.
      indent_level: (int) The indentation level.
    """
    if style.Get('USE_TABS'):
      if newlines_before > 0:
        indent_before = '\t' * indent_level + _TabbedContinuationAlignPadding(
            spaces, style.Get('CONTINUATION_ALIGN_STYLE'),
            style.Get('INDENT_WIDTH'))
      else:
        indent_before = '\t' * indent_level + ' ' * spaces
    else:
      indent_before = (' ' * indent_level * style.Get('INDENT_WIDTH') +
                       ' ' * spaces)

    if self.is_comment:
      comment_lines = [s.lstrip() for s in self.value.splitlines()]
      self.value = ('\n' + indent_before).join(comment_lines)

      # Update our own value since we are changing node value
      self.value = self.value

    if not self.whitespace_prefix:
      self.whitespace_prefix = ('\n' * (self.newlines or newlines_before) +
                                indent_before)
    else:
      self.whitespace_prefix += indent_before

  def AdjustNewlinesBefore(self, newlines_before):
    """Change the number of newlines before this token."""
    self.whitespace_prefix = ('\n' * newlines_before +
                              self.whitespace_prefix.lstrip('\n'))

  def RetainHorizontalSpacing(self, first_column, depth):
    """Retains a token's horizontal spacing."""
    previous = self.previous_token
    if not previous:
      return

    if previous.is_pseudo:
      previous = previous.previous_token
      if not previous:
        return

    cur_lineno = self.lineno
    prev_lineno = previous.lineno
    if previous.is_multiline_string:
      prev_lineno += previous.value.count('\n')

    if (cur_lineno != prev_lineno or
        (previous.is_pseudo and previous.value != ')' and
         cur_lineno != previous.previous_token.lineno)):
      self.spaces_required_before = (
          self.column - first_column + depth * style.Get('INDENT_WIDTH'))
      return

    cur_column = self.column
    prev_column = previous.column
    prev_len = len(previous.value)

    if previous.is_pseudo and previous.value == ')':
      prev_column -= 1
      prev_len = 0

    if previous.is_multiline_string:
      prev_len = len(previous.value.split('\n')[-1])
      if '\n' in previous.value:
        prev_column = 0  # Last line starts in column 0.

    self.spaces_required_before = cur_column - (prev_column + prev_len)

  def OpensScope(self):
    return self.value in _OPENING_BRACKETS

  def ClosesScope(self):
    return self.value in _CLOSING_BRACKETS

  def AddSubtype(self, subtype):
    self.subtypes.add(subtype)

  def __repr__(self):
    msg = ('FormatToken(name={0}, value={1}, column={2}, lineno={3}, '
           'splitpenalty={4}'.format(
               'DOCSTRING' if self.is_docstring else self.name, self.value,
               self.column, self.lineno, self.split_penalty))
    msg += ', pseudo)' if self.is_pseudo else ')'
    return msg

  @property
  def node_split_penalty(self):
    """Split penalty attached to the pytree node of this token."""
    return pytree_utils.GetNodeAnnotation(
        self.node, pytree_utils.Annotation.SPLIT_PENALTY, default=0)

  @property
  def is_binary_op(self):
    """Token is a binary operator."""
    return subtypes.BINARY_OPERATOR in self.subtypes

  @property
  @lru_cache()
  def is_arithmetic_op(self):
    """Token is an arithmetic operator."""
    return self.value in frozenset({
        '+',  # Add
        '-',  # Subtract
        '*',  # Multiply
        '@',  # Matrix Multiply
        '/',  # Divide
        '//',  # Floor Divide
        '%',  # Modulo
        '<<',  # Left Shift
        '>>',  # Right Shift
        '|',  # Bitwise Or
        '&',  # Bitwise Add
        '^',  # Bitwise Xor
        '**',  # Power
    })

  @property
  def is_simple_expr(self):
    """Token is an operator in a simple expression."""
    return subtypes.SIMPLE_EXPRESSION in self.subtypes

  @property
  def is_subscript_colon(self):
    """Token is a subscript colon."""
    return subtypes.SUBSCRIPT_COLON in self.subtypes

  @property
  def is_comment(self):
    return self.type == token.COMMENT

  @property
  def is_continuation(self):
    return self.type == CONTINUATION

  @property
  @lru_cache()
  def is_keyword(self):
    return keyword.iskeyword(
        self.value) or (self.value == 'match' and
                        type_repr(self.node.parent.type) == 'match_stmt') or (
                            self.value == 'case' and
                            type_repr(self.node.parent.type) == 'case_block')

  @property
  def is_name(self):
    return self.type == token.NAME and not self.is_keyword

  @property
  def is_number(self):
    return self.type == token.NUMBER

  @property
  def is_string(self):
    return self.type == token.STRING

  @property
  def is_multiline_string(self):
    """Test if this string is a multiline string.

    Returns:
      A multiline string always ends with triple quotes, so if it is a string
      token, inspect the last 3 characters and return True if it is a triple
      double or triple single quote mark.
    """
    return self.is_string and self.value.endswith(('"""', "'''"))

  @property
  def is_docstring(self):
    return self.is_string and self.previous_token is None

  @property
  def is_pylint_comment(self):
    return self.is_comment and re.match(r'#.*\bpylint:\s*(disable|enable)=',
                                        self.value)

  @property
  def is_pytype_comment(self):
    return self.is_comment and re.match(r'#.*\bpytype:\s*(disable|enable)=',
                                        self.value)

  @property
  def is_copybara_comment(self):
    return self.is_comment and re.match(
        r'#.*\bcopybara:\s*(strip|insert|replace)', self.value)



# Copyright 2006 Google, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
"""Pattern compiler.

The grammar is taken from PatternGrammar.txt.

The compiler compiles a pattern to a pytree.*Pattern instance.
"""

__author__ = 'Guido van Rossum <guido@python.org>'

# Python imports
import io

# Really local imports
from . import pygram
from . import pytree
# Fairly local imports
from .pgen2 import driver
from .pgen2 import grammar
from .pgen2 import literals
from .pgen2 import parse
from .pgen2 import token
from .pgen2 import tokenize


class PatternSyntaxError(Exception):
  pass


def tokenize_wrapper(input):
  """Tokenizes a string suppressing significant whitespace."""
  skip = {token.NEWLINE, token.INDENT, token.DEDENT}
  tokens = tokenize.generate_tokens(io.StringIO(input).readline)
  for quintuple in tokens:
    type, value, start, end, line_text = quintuple
    if type not in skip:
      yield quintuple


class PatternCompiler(object):

  def __init__(self, grammar_file=None):
    """Initializer.

        Takes an optional alternative filename for the pattern grammar.
        """
    if grammar_file is None:
      self.grammar = pygram.pattern_grammar
      self.syms = pygram.pattern_symbols
    else:
      self.grammar = driver.load_grammar(grammar_file)
      self.syms = pygram.Symbols(self.grammar)
    self.pygrammar = pygram.python_grammar
    self.pysyms = pygram.python_symbols
    self.driver = driver.Driver(self.grammar, convert=pattern_convert)

  def compile_pattern(self, input, debug=False, with_tree=False):
    """Compiles a pattern string to a nested pytree.*Pattern object."""
    tokens = tokenize_wrapper(input)
    try:
      root = self.driver.parse_tokens(tokens, debug=debug)
    except parse.ParseError as e:
      raise PatternSyntaxError(str(e)) from None
    if with_tree:
      return self.compile_node(root), root
    else:
      return self.compile_node(root)

  def compile_node(self, node):
    """Compiles a node, recursively.

        This is one big switch on the node type.
        """
    # XXX Optimize certain Wildcard-containing-Wildcard patterns
    # that can be merged
    if node.type == self.syms.Matcher:
      node = node.children[0]  # Avoid unneeded recursion

    if node.type == self.syms.Alternatives:
      # Skip the odd children since they are just '|' tokens
      alts = [self.compile_node(ch) for ch in node.children[::2]]
      if len(alts) == 1:
        return alts[0]
      p = pytree.WildcardPattern([[a] for a in alts], min=1, max=1)
      return p.optimize()

    if node.type == self.syms.Alternative:
      units = [self.compile_node(ch) for ch in node.children]
      if len(units) == 1:
        return units[0]
      p = pytree.WildcardPattern([units], min=1, max=1)
      return p.optimize()

    if node.type == self.syms.NegatedUnit:
      pattern = self.compile_basic(node.children[1:])
      p = pytree.NegatedPattern(pattern)
      return p.optimize()

    assert node.type == self.syms.Unit

    name = None
    nodes = node.children
    if len(nodes) >= 3 and nodes[1].type == token.EQUAL:
      name = nodes[0].value
      nodes = nodes[2:]
    repeat = None
    if len(nodes) >= 2 and nodes[-1].type == self.syms.Repeater:
      repeat = nodes[-1]
      nodes = nodes[:-1]

    # Now we've reduced it to: STRING | NAME [Details] | (...) | [...]
    pattern = self.compile_basic(nodes, repeat)

    if repeat is not None:
      assert repeat.type == self.syms.Repeater
      children = repeat.children
      child = children[0]
      if child.type == token.STAR:
        min = 0
        max = pytree.HUGE
      elif child.type == token.PLUS:
        min = 1
        max = pytree.HUGE
      elif child.type == token.LBRACE:
        assert children[-1].type == token.RBRACE
        assert len(children) in (3, 5)
        min = max = self.get_int(children[1])
        if len(children) == 5:
          max = self.get_int(children[3])
      else:
        assert False
      if min != 1 or max != 1:
        pattern = pattern.optimize()
        pattern = pytree.WildcardPattern([[pattern]], min=min, max=max)

    if name is not None:
      pattern.name = name
    return pattern.optimize()

  def compile_basic(self, nodes, repeat=None):
    # Compile STRING | NAME [Details] | (...) | [...]
    assert len(nodes) >= 1
    node = nodes[0]
    if node.type == token.STRING:
      value = str(literals.evalString(node.value))
      return pytree.LeafPattern(_type_of_literal(value), value)
    elif node.type == token.NAME:
      value = node.value
      if value.isupper():
        if value not in TOKEN_MAP:
          raise PatternSyntaxError('Invalid token: %r' % value)
        if nodes[1:]:
          raise PatternSyntaxError("Can't have details for token")
        return pytree.LeafPattern(TOKEN_MAP[value])
      else:
        if value == 'any':
          type = None
        elif not value.startswith('_'):
          type = getattr(self.pysyms, value, None)
          if type is None:
            raise PatternSyntaxError('Invalid symbol: %r' % value)
        if nodes[1:]:  # Details present
          content = [self.compile_node(nodes[1].children[1])]
        else:
          content = None
        return pytree.NodePattern(type, content)
    elif node.value == '(':
      return self.compile_node(nodes[1])
    elif node.value == '[':
      assert repeat is None
      subpattern = self.compile_node(nodes[1])
      return pytree.WildcardPattern([[subpattern]], min=0, max=1)
    assert False, node

  def get_int(self, node):
    assert node.type == token.NUMBER
    return int(node.value)


# Map named tokens to the type value for a LeafPattern
TOKEN_MAP = {
    'NAME': token.NAME,
    'STRING': token.STRING,
    'NUMBER': token.NUMBER,
    'TOKEN': None
}


def _type_of_literal(value):
  if value[0].isalpha():
    return token.NAME
  elif value in grammar.opmap:
    return grammar.opmap[value]
  else:
    return None


def pattern_convert(grammar, raw_node_info):
  """Converts raw node information to a Node or Leaf instance."""
  type, value, context, children = raw_node_info
  if children or type in grammar.number2symbol:
    return pytree.Node(type, children, context=context)
  else:
    return pytree.Leaf(type, value, context=context)


def compile_pattern(pattern):
  return PatternCompiler().compile_pattern(pattern)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python formatting style settings."""

import os
import re
import sys
import textwrap
from configparser import ConfigParser

from yapf.yapflib import errors

if sys.version_info >= (3, 11):
  import tomllib
else:
  import tomli as tomllib


class StyleConfigError(errors.YapfError):
  """Raised when there's a problem reading the style configuration."""
  pass


def Get(setting_name):
  """Get a style setting."""
  return _style[setting_name]


def GetOrDefault(setting_name, default_value):
  """Get a style setting or default value if the setting does not exist."""
  return _style.get(setting_name, default_value)


def Help():
  """Return dict mapping style names to help strings."""
  return _STYLE_HELP


def SetGlobalStyle(style):
  """Set a style dict."""
  global _style
  global _GLOBAL_STYLE_FACTORY
  factory = _GetStyleFactory(style)
  if factory:
    _GLOBAL_STYLE_FACTORY = factory
  _style = style


_STYLE_HELP = dict(
    # BASED_ON_STYLE='Which predefined style this style is based on',
    ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT=textwrap.dedent("""\
      Align closing bracket with visual indentation.
    """),
    ALLOW_MULTILINE_DICTIONARY_KEYS=textwrap.dedent("""\
      Allow dictionary keys to exist on multiple lines. For example:

        x = {
            ('this is the first element of a tuple',
             'this is the second element of a tuple'):
                 value,
        }
    """),
    ALLOW_MULTILINE_LAMBDAS=textwrap.dedent("""\
      Allow lambdas to be formatted on more than one line.
    """),
    ALLOW_SPLIT_BEFORE_DEFAULT_OR_NAMED_ASSIGNS=textwrap.dedent("""\
      Allow splitting before a default / named assignment in an argument list.
    """),
    ALLOW_SPLIT_BEFORE_DICT_VALUE=textwrap.dedent("""\
      Allow splits before the dictionary value.
    """),
    ARITHMETIC_PRECEDENCE_INDICATION=textwrap.dedent("""\
      Let spacing indicate operator precedence. For example:

        a = 1 * 2 + 3 / 4
        b = 1 / 2 - 3 * 4
        c = (1 + 2) * (3 - 4)
        d = (1 - 2) / (3 + 4)
        e = 1 * 2 - 3
        f = 1 + 2 + 3 + 4

    will be formatted as follows to indicate precedence:

        a = 1*2 + 3/4
        b = 1/2 - 3*4
        c = (1+2) * (3-4)
        d = (1-2) / (3+4)
        e = 1*2 - 3
        f = 1 + 2 + 3 + 4

    """),
    BLANK_LINE_BEFORE_CLASS_DOCSTRING=textwrap.dedent("""\
      Insert a blank line before a class-level docstring.
    """),
    BLANK_LINE_BEFORE_MODULE_DOCSTRING=textwrap.dedent("""\
      Insert a blank line before a module docstring.
    """),
    BLANK_LINE_BEFORE_NESTED_CLASS_OR_DEF=textwrap.dedent("""\
      Insert a blank line before a 'def' or 'class' immediately nested
      within another 'def' or 'class'. For example:

        class Foo:
                           # <------ this blank line
          def method():
            pass
    """),
    BLANK_LINES_AROUND_TOP_LEVEL_DEFINITION=textwrap.dedent("""\
      Number of blank lines surrounding top-level function and class
      definitions.
    """),
    BLANK_LINES_BETWEEN_TOP_LEVEL_IMPORTS_AND_VARIABLES=textwrap.dedent("""\
      Number of blank lines between top-level imports and variable
      definitions.
    """),
    COALESCE_BRACKETS=textwrap.dedent("""\
      Do not split consecutive brackets. Only relevant when
      dedent_closing_brackets is set. For example:

         call_func_that_takes_a_dict(
             {
                 'key1': 'value1',
                 'key2': 'value2',
             }
         )

      would reformat to:

         call_func_that_takes_a_dict({
             'key1': 'value1',
             'key2': 'value2',
         })
    """),
    COLUMN_LIMIT=textwrap.dedent("""\
      The column limit.
    """),
    CONTINUATION_ALIGN_STYLE=textwrap.dedent("""\
      The style for continuation alignment. Possible values are:

      - SPACE: Use spaces for continuation alignment. This is default behavior.
      - FIXED: Use fixed number (CONTINUATION_INDENT_WIDTH) of columns
        (ie: CONTINUATION_INDENT_WIDTH/INDENT_WIDTH tabs or
        CONTINUATION_INDENT_WIDTH spaces) for continuation alignment.
      - VALIGN-RIGHT: Vertically align continuation lines to multiple of
        INDENT_WIDTH columns. Slightly right (one tab or a few spaces) if
        cannot vertically align continuation lines with indent characters.
    """),
    CONTINUATION_INDENT_WIDTH=textwrap.dedent("""\
      Indent width used for line continuations.
    """),
    DEDENT_CLOSING_BRACKETS=textwrap.dedent("""\
      Put closing brackets on a separate line, dedented, if the bracketed
      expression can't fit in a single line. Applies to all kinds of brackets,
      including function definitions and calls. For example:

        config = {
            'key1': 'value1',
            'key2': 'value2',
        }        # <--- this bracket is dedented and on a separate line

        time_series = self.remote_client.query_entity_counters(
            entity='dev3246.region1',
            key='dns.query_latency_tcp',
            transform=Transformation.AVERAGE(window=timedelta(seconds=60)),
            start_ts=now()-timedelta(days=3),
            end_ts=now(),
        )        # <--- this bracket is dedented and on a separate line
    """),
    DISABLE_ENDING_COMMA_HEURISTIC=textwrap.dedent("""\
      Disable the heuristic which places each list element on a separate line
      if the list is comma-terminated.

      Note: The behavior of this flag changed in v0.40.3.  Before, if this flag
      was true, we would split lists that contained a trailing comma or a
      comment.  Now, we have a separate flag, `DISABLE_SPLIT_LIT_WITH_COMMENT`,
      that controls splitting when a list contains a comment.  To get the old
      behavior, set both flags to true.  More information in CHANGELOG.md.
    """),
    DISABLE_SPLIT_LIST_WITH_COMMENT=textwrap.dedent("""
      Don't put every element on a new line within a list that contains
      interstitial comments.
    """),
    EACH_DICT_ENTRY_ON_SEPARATE_LINE=textwrap.dedent("""\
      Place each dictionary entry onto its own line.
    """),
    FORCE_MULTILINE_DICT=textwrap.dedent("""\
      Require multiline dictionary even if it would normally fit on one line.
      For example:

        config = {
            'key1': 'value1'
        }
    """),
    I18N_COMMENT=textwrap.dedent("""\
      The regex for an i18n comment. The presence of this comment stops
      reformatting of that line, because the comments are required to be
      next to the string they translate.
    """),
    I18N_FUNCTION_CALL=textwrap.dedent("""\
      The i18n function call names. The presence of this function stops
      reformattting on that line, because the string it has cannot be moved
      away from the i18n comment.
    """),
    INDENT_CLOSING_BRACKETS=textwrap.dedent("""\
      Put closing brackets on a separate line, indented, if the bracketed
      expression can't fit in a single line. Applies to all kinds of brackets,
      including function definitions and calls. For example:

        config = {
            'key1': 'value1',
            'key2': 'value2',
            }        # <--- this bracket is indented and on a separate line

        time_series = self.remote_client.query_entity_counters(
            entity='dev3246.region1',
            key='dns.query_latency_tcp',
            transform=Transformation.AVERAGE(window=timedelta(seconds=60)),
            start_ts=now()-timedelta(days=3),
            end_ts=now(),
            )        # <--- this bracket is indented and on a separate line
    """),
    INDENT_DICTIONARY_VALUE=textwrap.dedent("""\
      Indent the dictionary value if it cannot fit on the same line as the
      dictionary key. For example:

        config = {
            'key1':
                'value1',
            'key2': value1 +
                    value2,
        }
    """),
    INDENT_BLANK_LINES=textwrap.dedent("""\
      Indent blank lines.
    """),
    INDENT_WIDTH=textwrap.dedent("""\
      The number of columns to use for indentation.
    """),
    JOIN_MULTIPLE_LINES=textwrap.dedent("""\
      Join short lines into one line. E.g., single line 'if' statements.
    """),
    NO_SPACES_AROUND_SELECTED_BINARY_OPERATORS=textwrap.dedent("""\
      Do not include spaces around selected binary operators. For example:

        1 + 2 * 3 - 4 / 5

      will be formatted as follows when configured with "*,/":

        1 + 2*3 - 4/5
    """),
    SPACE_BETWEEN_ENDING_COMMA_AND_CLOSING_BRACKET=textwrap.dedent("""\
      Insert a space between the ending comma and closing bracket of a list,
      etc.
    """),
    SPACE_INSIDE_BRACKETS=textwrap.dedent("""\
      Use spaces inside brackets, braces, and parentheses.  For example:

        method_call( 1 )
        my_dict[ 3 ][ 1 ][ get_index( *args, **kwargs ) ]
        my_set = { 1, 2, 3 }
    """),
    SPACES_AROUND_DEFAULT_OR_NAMED_ASSIGN=textwrap.dedent("""\
      Use spaces around default or named assigns.
    """),
    SPACES_AROUND_DICT_DELIMITERS=textwrap.dedent("""\
      Adds a space after the opening '{' and before the ending '}' dict
      delimiters.

        {1: 2}

      will be formatted as:

        { 1: 2 }
    """),
    SPACES_AROUND_LIST_DELIMITERS=textwrap.dedent("""\
      Adds a space after the opening '[' and before the ending ']' list
      delimiters.

        [1, 2]

      will be formatted as:

        [ 1, 2 ]
    """),
    SPACES_AROUND_POWER_OPERATOR=textwrap.dedent("""\
      Use spaces around the power operator.
    """),
    SPACES_AROUND_SUBSCRIPT_COLON=textwrap.dedent("""\
      Use spaces around the subscript / slice operator.  For example:

        my_list[1 : 10 : 2]
    """),
    SPACES_AROUND_TUPLE_DELIMITERS=textwrap.dedent("""\
      Adds a space after the opening '(' and before the ending ')' tuple
      delimiters.

        (1, 2, 3)

      will be formatted as:

        ( 1, 2, 3 )
    """),
    SPACES_BEFORE_COMMENT=textwrap.dedent("""\
      The number of spaces required before a trailing comment.
      This can be a single value (representing the number of spaces
      before each trailing comment) or list of values (representing
      alignment column values; trailing comments within a block will
      be aligned to the first column value that is greater than the maximum
      line length within the block). For example:

      With spaces_before_comment=5:

        1 + 1 # Adding values

      will be formatted as:

        1 + 1     # Adding values <-- 5 spaces between the end of the
                  # statement and comment

      With spaces_before_comment=15, 20:

        1 + 1 # Adding values
        two + two # More adding

        longer_statement # This is a longer statement
        short # This is a shorter statement

        a_very_long_statement_that_extends_beyond_the_final_column # Comment
        short # This is a shorter statement

      will be formatted as:

        1 + 1          # Adding values <-- end of line comments in block
                       # aligned to col 15
        two + two      # More adding

        longer_statement    # This is a longer statement <-- end of line
                            # comments in block aligned to col 20
        short               # This is a shorter statement

        a_very_long_statement_that_extends_beyond_the_final_column  # Comment <-- the end of line comments are aligned based on the line length
        short                                                       # This is a shorter statement

    """),  # noqa
    SPLIT_ALL_COMMA_SEPARATED_VALUES=textwrap.dedent("""\
      Split before arguments.
    """),
    SPLIT_ALL_TOP_LEVEL_COMMA_SEPARATED_VALUES=textwrap.dedent("""\
      Split before arguments, but do not split all subexpressions recursively
      (unless needed).
    """),
    SPLIT_ARGUMENTS_WHEN_COMMA_TERMINATED=textwrap.dedent("""\
      Split before arguments if the argument list is terminated by a
      comma.
    """),
    SPLIT_BEFORE_ARITHMETIC_OPERATOR=textwrap.dedent("""\
      Set to True to prefer splitting before '+', '-', '*', '/', '//', or '@'
      rather than after.
    """),
    SPLIT_BEFORE_BITWISE_OPERATOR=textwrap.dedent("""\
      Set to True to prefer splitting before '&', '|' or '^' rather than
      after.
    """),
    SPLIT_BEFORE_CLOSING_BRACKET=textwrap.dedent("""\
      Split before the closing bracket if a list or dict literal doesn't fit on
      a single line.
    """),
    SPLIT_BEFORE_DICT_SET_GENERATOR=textwrap.dedent("""\
      Split before a dictionary or set generator (comp_for). For example, note
      the split before the 'for':

        foo = {
            variable: 'Hello world, have a nice day!'
            for variable in bar if variable != 42
        }
    """),
    SPLIT_BEFORE_DOT=textwrap.dedent("""\
      Split before the '.' if we need to split a longer expression:

        foo = ('This is a really long string: {}, {}, {}, {}'.format(a, b, c, d))

      would reformat to something like:

        foo = ('This is a really long string: {}, {}, {}, {}'
               .format(a, b, c, d))
    """),  # noqa
    SPLIT_BEFORE_EXPRESSION_AFTER_OPENING_PAREN=textwrap.dedent("""\
      Split after the opening paren which surrounds an expression if it doesn't
      fit on a single line.
    """),
    SPLIT_BEFORE_FIRST_ARGUMENT=textwrap.dedent("""\
      If an argument / parameter list is going to be split, then split before
      the first argument.
    """),
    SPLIT_BEFORE_LOGICAL_OPERATOR=textwrap.dedent("""\
      Set to True to prefer splitting before 'and' or 'or' rather than
      after.
    """),
    SPLIT_BEFORE_NAMED_ASSIGNS=textwrap.dedent("""\
      Split named assignments onto individual lines.
    """),
    SPLIT_COMPLEX_COMPREHENSION=textwrap.dedent("""\
      Set to True to split list comprehensions and generators that have
      non-trivial expressions and multiple clauses before each of these
      clauses. For example:

        result = [
            a_long_var + 100 for a_long_var in xrange(1000)
            if a_long_var % 10]

      would reformat to something like:

        result = [
            a_long_var + 100
            for a_long_var in xrange(1000)
            if a_long_var % 10]
    """),
    SPLIT_PENALTY_AFTER_OPENING_BRACKET=textwrap.dedent("""\
      The penalty for splitting right after the opening bracket.
    """),
    SPLIT_PENALTY_AFTER_UNARY_OPERATOR=textwrap.dedent("""\
      The penalty for splitting the line after a unary operator.
    """),
    SPLIT_PENALTY_ARITHMETIC_OPERATOR=textwrap.dedent("""\
      The penalty of splitting the line around the '+', '-', '*', '/', '//',
      `%`, and '@' operators.
    """),
    SPLIT_PENALTY_BEFORE_IF_EXPR=textwrap.dedent("""\
      The penalty for splitting right before an if expression.
    """),
    SPLIT_PENALTY_BITWISE_OPERATOR=textwrap.dedent("""\
      The penalty of splitting the line around the '&', '|', and '^' operators.
    """),
    SPLIT_PENALTY_COMPREHENSION=textwrap.dedent("""\
      The penalty for splitting a list comprehension or generator
      expression.
    """),
    SPLIT_PENALTY_EXCESS_CHARACTER=textwrap.dedent("""\
      The penalty for characters over the column limit.
    """),
    SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT=textwrap.dedent("""\
      The penalty incurred by adding a line split to the logical line. The
      more line splits added the higher the penalty.
    """),
    SPLIT_PENALTY_IMPORT_NAMES=textwrap.dedent("""\
      The penalty of splitting a list of "import as" names. For example:

        from a_very_long_or_indented_module_name_yada_yad import (long_argument_1,
                                                                  long_argument_2,
                                                                  long_argument_3)

      would reformat to something like:

        from a_very_long_or_indented_module_name_yada_yad import (
            long_argument_1, long_argument_2, long_argument_3)
    """),  # noqa
    SPLIT_PENALTY_LOGICAL_OPERATOR=textwrap.dedent("""\
      The penalty of splitting the line around the 'and' and 'or' operators.
    """),
    USE_TABS=textwrap.dedent("""\
      Use the Tab character for indentation.
    """),
)


def CreatePEP8Style():
  """Create the PEP8 formatting style."""
  return dict(
      ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT=True,
      ALLOW_MULTILINE_DICTIONARY_KEYS=False,
      ALLOW_MULTILINE_LAMBDAS=False,
      ALLOW_SPLIT_BEFORE_DEFAULT_OR_NAMED_ASSIGNS=True,
      ALLOW_SPLIT_BEFORE_DICT_VALUE=True,
      ARITHMETIC_PRECEDENCE_INDICATION=False,
      BLANK_LINE_BEFORE_CLASS_DOCSTRING=False,
      BLANK_LINE_BEFORE_MODULE_DOCSTRING=False,
      BLANK_LINE_BEFORE_NESTED_CLASS_OR_DEF=True,
      BLANK_LINES_AROUND_TOP_LEVEL_DEFINITION=2,
      BLANK_LINES_BETWEEN_TOP_LEVEL_IMPORTS_AND_VARIABLES=1,
      COALESCE_BRACKETS=False,
      COLUMN_LIMIT=79,
      CONTINUATION_ALIGN_STYLE='SPACE',
      CONTINUATION_INDENT_WIDTH=4,
      DEDENT_CLOSING_BRACKETS=False,
      DISABLE_ENDING_COMMA_HEURISTIC=False,
      DISABLE_SPLIT_LIST_WITH_COMMENT=False,
      EACH_DICT_ENTRY_ON_SEPARATE_LINE=True,
      FORCE_MULTILINE_DICT=False,
      I18N_COMMENT='',
      I18N_FUNCTION_CALL='',
      INDENT_CLOSING_BRACKETS=False,
      INDENT_DICTIONARY_VALUE=False,
      INDENT_WIDTH=4,
      INDENT_BLANK_LINES=False,
      JOIN_MULTIPLE_LINES=True,
      NO_SPACES_AROUND_SELECTED_BINARY_OPERATORS=set(),
      SPACE_BETWEEN_ENDING_COMMA_AND_CLOSING_BRACKET=True,
      SPACE_INSIDE_BRACKETS=False,
      SPACES_AROUND_DEFAULT_OR_NAMED_ASSIGN=False,
      SPACES_AROUND_DICT_DELIMITERS=False,
      SPACES_AROUND_LIST_DELIMITERS=False,
      SPACES_AROUND_POWER_OPERATOR=False,
      SPACES_AROUND_SUBSCRIPT_COLON=False,
      SPACES_AROUND_TUPLE_DELIMITERS=False,
      SPACES_BEFORE_COMMENT=2,
      SPLIT_ALL_COMMA_SEPARATED_VALUES=False,
      SPLIT_ALL_TOP_LEVEL_COMMA_SEPARATED_VALUES=False,
      SPLIT_ARGUMENTS_WHEN_COMMA_TERMINATED=False,
      SPLIT_BEFORE_ARITHMETIC_OPERATOR=False,
      SPLIT_BEFORE_BITWISE_OPERATOR=True,
      SPLIT_BEFORE_CLOSING_BRACKET=True,
      SPLIT_BEFORE_DICT_SET_GENERATOR=True,
      SPLIT_BEFORE_DOT=False,
      SPLIT_BEFORE_EXPRESSION_AFTER_OPENING_PAREN=False,
      SPLIT_BEFORE_FIRST_ARGUMENT=False,
      SPLIT_BEFORE_LOGICAL_OPERATOR=True,
      SPLIT_BEFORE_NAMED_ASSIGNS=True,
      SPLIT_COMPLEX_COMPREHENSION=False,
      SPLIT_PENALTY_AFTER_OPENING_BRACKET=300,
      SPLIT_PENALTY_AFTER_UNARY_OPERATOR=10000,
      SPLIT_PENALTY_ARITHMETIC_OPERATOR=300,
      SPLIT_PENALTY_BEFORE_IF_EXPR=0,
      SPLIT_PENALTY_BITWISE_OPERATOR=300,
      SPLIT_PENALTY_COMPREHENSION=80,
      SPLIT_PENALTY_EXCESS_CHARACTER=7000,
      SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT=30,
      SPLIT_PENALTY_IMPORT_NAMES=0,
      SPLIT_PENALTY_LOGICAL_OPERATOR=300,
      USE_TABS=False,
  )


def CreateGoogleStyle():
  """Create the Google formatting style."""
  style = CreatePEP8Style()
  style['ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT'] = False
  style['COLUMN_LIMIT'] = 80
  style['INDENT_DICTIONARY_VALUE'] = True
  style['INDENT_WIDTH'] = 4
  style['I18N_COMMENT'] = r'#\..*'
  style['I18N_FUNCTION_CALL'] = ['N_', '_']
  style['JOIN_MULTIPLE_LINES'] = False
  style['SPACE_BETWEEN_ENDING_COMMA_AND_CLOSING_BRACKET'] = False
  style['SPLIT_BEFORE_BITWISE_OPERATOR'] = False
  style['SPLIT_BEFORE_DICT_SET_GENERATOR'] = False
  style['SPLIT_BEFORE_LOGICAL_OPERATOR'] = False
  style['SPLIT_COMPLEX_COMPREHENSION'] = True
  style['SPLIT_PENALTY_COMPREHENSION'] = 2100
  return style


def CreateYapfStyle():
  """Create the YAPF formatting style."""
  style = CreateGoogleStyle()
  style['ALLOW_MULTILINE_DICTIONARY_KEYS'] = True
  style['ALLOW_SPLIT_BEFORE_DEFAULT_OR_NAMED_ASSIGNS'] = False
  style['INDENT_WIDTH'] = 2
  style['SPLIT_BEFORE_BITWISE_OPERATOR'] = True
  style['SPLIT_BEFORE_DOT'] = True
  style['SPLIT_BEFORE_EXPRESSION_AFTER_OPENING_PAREN'] = True
  return style


def CreateFacebookStyle():
  """Create the Facebook formatting style."""
  style = CreatePEP8Style()
  style['ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT'] = False
  style['BLANK_LINE_BEFORE_NESTED_CLASS_OR_DEF'] = False
  style['COLUMN_LIMIT'] = 80
  style['DEDENT_CLOSING_BRACKETS'] = True
  style['INDENT_CLOSING_BRACKETS'] = False
  style['INDENT_DICTIONARY_VALUE'] = True
  style['JOIN_MULTIPLE_LINES'] = False
  style['SPACES_BEFORE_COMMENT'] = 2
  style['SPLIT_PENALTY_AFTER_OPENING_BRACKET'] = 0
  style['SPLIT_PENALTY_BEFORE_IF_EXPR'] = 30
  style['SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT'] = 30
  style['SPLIT_BEFORE_BITWISE_OPERATOR'] = False
  style['SPLIT_BEFORE_LOGICAL_OPERATOR'] = False
  return style


_STYLE_NAME_TO_FACTORY = dict(
    facebook=CreateFacebookStyle,
    google=CreateGoogleStyle,
    pep8=CreatePEP8Style,
    yapf=CreateYapfStyle,
)

_DEFAULT_STYLE_TO_FACTORY = [
    (CreateFacebookStyle(), CreateFacebookStyle),
    (CreateGoogleStyle(), CreateGoogleStyle),
    (CreatePEP8Style(), CreatePEP8Style),
    (CreateYapfStyle(), CreateYapfStyle),
]


def _GetStyleFactory(style):
  for def_style, factory in _DEFAULT_STYLE_TO_FACTORY:
    if style == def_style:
      return factory
  return None


def _ContinuationAlignStyleStringConverter(s):
  """Option value converter for a continuation align style string."""
  accepted_styles = ('SPACE', 'FIXED', 'VALIGN-RIGHT')
  if s:
    r = s.strip('"\'').replace('_', '-').upper()
    if r not in accepted_styles:
      raise ValueError('unknown continuation align style: %r' % (s,))
  else:
    r = accepted_styles[0]
  return r


def _StringListConverter(s):
  """Option value converter for a comma-separated list of strings."""
  return [part.strip() for part in s.split(',')]


def _StringSetConverter(s):
  """Option value converter for a comma-separated set of strings."""
  if len(s) > 2 and s[0] in '"\'':
    s = s[1:-1]
  return {part.strip() for part in s.split(',')}


def _BoolConverter(s):
  """Option value converter for a boolean."""
  return ConfigParser.BOOLEAN_STATES[s.lower()]


def _IntListConverter(s):
  """Option value converter for a comma-separated list of integers."""
  s = s.strip()
  if s.startswith('[') and s.endswith(']'):
    s = s[1:-1]

  return [int(part.strip()) for part in s.split(',') if part.strip()]


def _IntOrIntListConverter(s):
  """Option value converter for an integer or list of integers."""
  if len(s) > 2 and s[0] in '"\'':
    s = s[1:-1]
  return _IntListConverter(s) if ',' in s else int(s)


# Different style options need to have their values interpreted differently when
# read from the config file. This dict maps an option name to a "converter"
# function that accepts the string read for the option's value from the file and
# returns it wrapper in actual Python type that's going to be meaningful to
# yapf.
#
# Note: this dict has to map all the supported style options.
_STYLE_OPTION_VALUE_CONVERTER = dict(
    ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT=_BoolConverter,
    ALLOW_MULTILINE_DICTIONARY_KEYS=_BoolConverter,
    ALLOW_MULTILINE_LAMBDAS=_BoolConverter,
    ALLOW_SPLIT_BEFORE_DEFAULT_OR_NAMED_ASSIGNS=_BoolConverter,
    ALLOW_SPLIT_BEFORE_DICT_VALUE=_BoolConverter,
    ARITHMETIC_PRECEDENCE_INDICATION=_BoolConverter,
    BLANK_LINE_BEFORE_CLASS_DOCSTRING=_BoolConverter,
    BLANK_LINE_BEFORE_MODULE_DOCSTRING=_BoolConverter,
    BLANK_LINE_BEFORE_NESTED_CLASS_OR_DEF=_BoolConverter,
    BLANK_LINES_AROUND_TOP_LEVEL_DEFINITION=int,
    BLANK_LINES_BETWEEN_TOP_LEVEL_IMPORTS_AND_VARIABLES=int,
    COALESCE_BRACKETS=_BoolConverter,
    COLUMN_LIMIT=int,
    CONTINUATION_ALIGN_STYLE=_ContinuationAlignStyleStringConverter,
    CONTINUATION_INDENT_WIDTH=int,
    DEDENT_CLOSING_BRACKETS=_BoolConverter,
    DISABLE_ENDING_COMMA_HEURISTIC=_BoolConverter,
    DISABLE_SPLIT_LIST_WITH_COMMENT=_BoolConverter,
    EACH_DICT_ENTRY_ON_SEPARATE_LINE=_BoolConverter,
    FORCE_MULTILINE_DICT=_BoolConverter,
    I18N_COMMENT=str,
    I18N_FUNCTION_CALL=_StringListConverter,
    INDENT_BLANK_LINES=_BoolConverter,
    INDENT_CLOSING_BRACKETS=_BoolConverter,
    INDENT_DICTIONARY_VALUE=_BoolConverter,
    INDENT_WIDTH=int,
    JOIN_MULTIPLE_LINES=_BoolConverter,
    NO_SPACES_AROUND_SELECTED_BINARY_OPERATORS=_StringSetConverter,
    SPACE_BETWEEN_ENDING_COMMA_AND_CLOSING_BRACKET=_BoolConverter,
    SPACE_INSIDE_BRACKETS=_BoolConverter,
    SPACES_AROUND_DEFAULT_OR_NAMED_ASSIGN=_BoolConverter,
    SPACES_AROUND_DICT_DELIMITERS=_BoolConverter,
    SPACES_AROUND_LIST_DELIMITERS=_BoolConverter,
    SPACES_AROUND_POWER_OPERATOR=_BoolConverter,
    SPACES_AROUND_SUBSCRIPT_COLON=_BoolConverter,
    SPACES_AROUND_TUPLE_DELIMITERS=_BoolConverter,
    SPACES_BEFORE_COMMENT=_IntOrIntListConverter,
    SPLIT_ALL_COMMA_SEPARATED_VALUES=_BoolConverter,
    SPLIT_ALL_TOP_LEVEL_COMMA_SEPARATED_VALUES=_BoolConverter,
    SPLIT_ARGUMENTS_WHEN_COMMA_TERMINATED=_BoolConverter,
    SPLIT_BEFORE_ARITHMETIC_OPERATOR=_BoolConverter,
    SPLIT_BEFORE_BITWISE_OPERATOR=_BoolConverter,
    SPLIT_BEFORE_CLOSING_BRACKET=_BoolConverter,
    SPLIT_BEFORE_DICT_SET_GENERATOR=_BoolConverter,
    SPLIT_BEFORE_DOT=_BoolConverter,
    SPLIT_BEFORE_EXPRESSION_AFTER_OPENING_PAREN=_BoolConverter,
    SPLIT_BEFORE_FIRST_ARGUMENT=_BoolConverter,
    SPLIT_BEFORE_LOGICAL_OPERATOR=_BoolConverter,
    SPLIT_BEFORE_NAMED_ASSIGNS=_BoolConverter,
    SPLIT_COMPLEX_COMPREHENSION=_BoolConverter,
    SPLIT_PENALTY_AFTER_OPENING_BRACKET=int,
    SPLIT_PENALTY_AFTER_UNARY_OPERATOR=int,
    SPLIT_PENALTY_ARITHMETIC_OPERATOR=int,
    SPLIT_PENALTY_BEFORE_IF_EXPR=int,
    SPLIT_PENALTY_BITWISE_OPERATOR=int,
    SPLIT_PENALTY_COMPREHENSION=int,
    SPLIT_PENALTY_EXCESS_CHARACTER=int,
    SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT=int,
    SPLIT_PENALTY_IMPORT_NAMES=int,
    SPLIT_PENALTY_LOGICAL_OPERATOR=int,
    USE_TABS=_BoolConverter,
)


def CreateStyleFromConfig(style_config):
  """Create a style dict from the given config.

  Arguments:
    style_config: either a style name or a file name. The file is expected to
      contain settings. It can have a special BASED_ON_STYLE setting naming the
      style which it derives from. If no such setting is found, it derives from
      the default style. When style_config is None, the _GLOBAL_STYLE_FACTORY
      config is created.

  Returns:
    A style dict.

  Raises:
    StyleConfigError: if an unknown style option was encountered.
  """

  def GlobalStyles():
    for style, _ in _DEFAULT_STYLE_TO_FACTORY:
      yield style

  def_style = False
  if style_config is None:
    for style in GlobalStyles():
      if _style == style:
        def_style = True
        break
    if not def_style:
      return _style
    return _GLOBAL_STYLE_FACTORY()

  if isinstance(style_config, dict):
    config = _CreateConfigParserFromConfigDict(style_config)
  elif isinstance(style_config, str):
    style_factory = _STYLE_NAME_TO_FACTORY.get(style_config.lower())
    if style_factory is not None:
      return style_factory()
    if style_config.startswith('{'):
      # Most likely a style specification from the command line.
      config = _CreateConfigParserFromConfigString(style_config)
    else:
      # Unknown config name: assume it's a file name then.
      config = _CreateConfigParserFromConfigFile(style_config)
  return _CreateStyleFromConfigParser(config)


def _CreateConfigParserFromConfigDict(config_dict):
  config = ConfigParser()
  config.add_section('style')
  for key, value in config_dict.items():
    config.set('style', key, str(value))
  return config


def _CreateConfigParserFromConfigString(config_string):
  """Given a config string from the command line, return a config parser."""
  if config_string[0] != '{' or config_string[-1] != '}':
    raise StyleConfigError(
        "Invalid style dict syntax: '{}'.".format(config_string))
  config = ConfigParser()
  config.add_section('style')
  for key, value, _ in re.findall(
      r'([a-zA-Z0-9_]+)\s*[:=]\s*'
      r'(?:'
      r'((?P<quote>[\'"]).*?(?P=quote)|'
      r'[a-zA-Z0-9_]+)'
      r')', config_string):  # yapf: disable
    config.set('style', key, value)
  return config


def _CreateConfigParserFromConfigFile(config_filename):
  """Read the file and return a ConfigParser object."""
  if not os.path.exists(config_filename):
    # Provide a more meaningful error here.
    raise StyleConfigError(
        '"{0}" is not a valid style or file path'.format(config_filename))
  config = ConfigParser()

  if config_filename.endswith(PYPROJECT_TOML):
    with open(config_filename, 'rb') as style_file:
      pyproject_toml = tomllib.load(style_file)
      style_dict = pyproject_toml.get('tool', {}).get('yapf', None)
      if style_dict is None:
        raise StyleConfigError(
            'Unable to find section [tool.yapf] in {0}'.format(config_filename))
      config.add_section('style')
      for k, v in style_dict.items():
        config.set('style', k, str(v))
      return config

  with open(config_filename) as style_file:
    config.read_file(style_file)

    if config_filename.endswith(SETUP_CONFIG):
      if not config.has_section('yapf'):
        raise StyleConfigError(
            'Unable to find section [yapf] in {0}'.format(config_filename))
      return config

    if config_filename.endswith(LOCAL_STYLE):
      if not config.has_section('style'):
        raise StyleConfigError(
            'Unable to find section [style] in {0}'.format(config_filename))
      return config

    if not config.has_section('style'):
      raise StyleConfigError(
          'Unable to find section [style] in {0}'.format(config_filename))
    return config


def _CreateStyleFromConfigParser(config):
  """Create a style dict from a configuration file.

  Arguments:
    config: a ConfigParser object.

  Returns:
    A style dict.

  Raises:
    StyleConfigError: if an unknown style option was encountered.
  """
  # Initialize the base style.
  section = 'yapf' if config.has_section('yapf') else 'style'
  if config.has_option('style', 'based_on_style'):
    based_on = config.get('style', 'based_on_style').lower()
    base_style = _STYLE_NAME_TO_FACTORY[based_on]()
  elif config.has_option('yapf', 'based_on_style'):
    based_on = config.get('yapf', 'based_on_style').lower()
    base_style = _STYLE_NAME_TO_FACTORY[based_on]()
  else:
    base_style = _GLOBAL_STYLE_FACTORY()

  # Read all options specified in the file and update the style.
  for option, value in config.items(section):
    if option.lower() == 'based_on_style':
      # Now skip this one - we've already handled it and it's not one of the
      # recognized style options.
      continue
    option = option.upper()
    if option not in _STYLE_OPTION_VALUE_CONVERTER:
      raise StyleConfigError('Unknown style option "{0}"'.format(option))
    try:
      base_style[option] = _STYLE_OPTION_VALUE_CONVERTER[option](value)
    except ValueError:
      raise StyleConfigError("'{}' is not a valid setting for {}.".format(
          value, option))
  return base_style


# The default style - used if yapf is not invoked without specifically
# requesting a formatting style.
DEFAULT_STYLE = 'pep8'
DEFAULT_STYLE_FACTORY = CreatePEP8Style
_GLOBAL_STYLE_FACTORY = CreatePEP8Style

# The name of the file to use for global style definition.
GLOBAL_STYLE = (
    os.path.join(
        os.getenv('XDG_CONFIG_HOME') or os.path.expanduser('~/.config'), 'yapf',
        'style'))

# The name of the file to use for directory-local style definition.
LOCAL_STYLE = '.style.yapf'

# Alternative place for directory-local style definition. Style should be
# specified in the '[yapf]' section.
SETUP_CONFIG = 'setup.cfg'

# Style definition by local pyproject.toml file. Style should be specified
# in the '[tool.yapf]' section.
PYPROJECT_TOML = 'pyproject.toml'

# TODO(eliben): For now we're preserving the global presence of a style dict.
# Refactor this so that the style is passed around through yapf rather than
# being global.
_style = None
SetGlobalStyle(_GLOBAL_STYLE_FACTORY())



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
"""Safely evaluate Python string literals without using eval()."""

import re

simple_escapes = {
    'a': '\a',
    'b': '\b',
    'f': '\f',
    'n': '\n',
    'r': '\r',
    't': '\t',
    'v': '\v',
    "'": "'",
    '"': '"',
    '\\': '\\'
}


def escape(m):
  all, tail = m.group(0, 1)
  assert all.startswith('\\')
  esc = simple_escapes.get(tail)
  if esc is not None:
    return esc
  if tail.startswith('x'):
    hexes = tail[1:]
    if len(hexes) < 2:
      raise ValueError("invalid hex string escape ('\\%s')" % tail)
    try:
      i = int(hexes, 16)
    except ValueError:
      raise ValueError("invalid hex string escape ('\\%s')" % tail) from None
  else:
    try:
      i = int(tail, 8)
    except ValueError:
      raise ValueError("invalid octal string escape ('\\%s')" % tail) from None
  return chr(i)


def evalString(s):
  assert s.startswith("'") or s.startswith('"'), repr(s[:1])
  q = s[0]
  if s[:3] == q * 3:
    q = q * 3
  assert s.endswith(q), repr(s[-len(q):])
  assert len(s) >= 2 * len(q)
  s = s[len(q):-len(q)]
  return re.sub(r"\\(\'|\"|\\|[abfnrtv]|x.{0,2}|[0-7]{1,3})", escape, s)


def test():
  for i in range(256):
    c = chr(i)
    s = repr(c)
    e = evalString(s)
    if e != c:
      print(i, c, s, e)


if __name__ == '__main__':
  test()
